'''
    Copyright (c) 2018-, S. Shiraiwa
    All Rights reserved. See file COPYRIGHT for details.

    Variables


    This modules interface string exression to MFEM

    for example, when a user write

       epsiolnr = 'x + 0.5y'

     and if epsilonr is one of Variable object, it will become

        call of epsilon(x,y,z) at integration points (matrix assembly)

        or

        many such calles for all nodal points (plot)

    about variable decorator:
       this class instance is used to convered a user written function
       to a Vriable object.

    from petram.helper.variables import variable

    @variable.float()
    def test(x, y, z):
       return 1-0.1j

    @variable.float(dependency = ("u",))
    def test(x, y, z):
       # u is FES variable solved in the previous space.
       value = u()
       return value

    @variable.complex()
    def ctest(x, y, z):
       return 1-0.1j

    @variable.array(complex=True,shape=(2,))
    def atest(x, y, z):
       return np.array([1-0.1j,1-0.1j])

'''
import numpy as np
import weakref
import types
import traceback
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD

from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Variables')


class _decorator(object):
    @staticmethod
    def float(func=None, *, dependency=None, grad=None, curl=None, div=None, td=False):
        '''
        this form allows for using both
        @float
        @float()
        '''
        def wrapper(func):
            def dec(*args, **kwargs):
                obj = PyFunctionVariable(func,
                                         complex=False,
                                         dependency=dependency,
                                         grad=grad,
                                         curl=curl,
                                         div=div)
                return obj
            return dec(func)
        if func:
            return wrapper(func)
        else:
            return wrapper

    @staticmethod
    def complex(func=None, *, dependency=None, grad=None, curl=None, div=None, td=False):
        '''
        this form allows for using both
        @complex
        @complex()
        '''
        def wrapper(func):
            def dec(*args, **kwargs):
                obj = PyFunctionVariable(func,
                                         complex=True,
                                         dependency=dependency,
                                         grad=grad,
                                         curl=curl,
                                         div=div)

                return obj
            return dec(func)
        if func:
            return wrapper(func)
        else:
            return wrapper

    @staticmethod
    def array(complex=False, shape=(1,), dependency=None, grad=None,
              curl=None, div=None, td=False):
        def dec(func):
            obj = PyFunctionVariable(func,
                                     complex=complex,
                                     shape=shape,
                                     dependency=dependency,
                                     grad=grad,
                                     curl=curl,
                                     div=div,)
            return obj
        return dec


class _decorator_jit(object):

    @staticmethod
    def float(func=None, *, dependency=None, grad=None, curl=None, div=None, td=False, params=None):
        '''
        This form allows to use both with and without ()
        @float
        @float()
        @float(dependency....
        '''
        def wrapper(func):
            def dec(*args, **kwargs):
                obj = NumbaCoefficientVariable(func,
                                               complex=False,
                                               dependency=dependency,
                                               params=params,
                                               grad=grad,
                                               curl=curl,
                                               div=div,
                                               td=td,)
                return obj
            return dec(func)
        if func:
            return wrapper(func)
        else:
            return wrapper

    @staticmethod
    def complex(func=None, *, dependency=None, grad=None, curl=None, div=None, td=False, params=None):
        '''
        This form allows to use both with and without ()
        @complex
        @complex()
        @complex(dependency....
        '''
        def wrapper(func):
            def dec(*args, **kwargs):
                obj = NumbaCoefficientVariable(func,
                                               complex=True,
                                               dependency=dependency,
                                               params=params,
                                               grad=grad,
                                               curl=curl,
                                               div=div,
                                               td=td,)
                return obj
            return dec(func)
        if func:
            return wrapper(func)
        else:
            return wrapper

    @staticmethod
    def array(complex=False, shape=(1,), dependency=None, grad=None, curl=None, div=None, td=False, params=None):
        def dec(func):
            obj = NumbaCoefficientVariable(func,
                                           complex=complex,
                                           shape=shape,
                                           dependency=dependency,
                                           params=params,
                                           grad=grad,
                                           curl=curl,
                                           div=div,
                                           td=td,)

            return obj
        return dec


variable = _decorator()
variable.jit = _decorator_jit()


def eval_code(co, g, l, flag=None):
    if flag is not None:
        if not flag:
            return co
    else:
        if not isinstance(co, types.CodeType):
            return co
    try:
        a = eval(co, g, l)
    except NameError:
        dprint1("global names", g.keys())
        dprint1("local names", l.keys())
        raise
    if callable(a):
        return a()
    return a


def cosd(x): return np.cos(x * np.pi / 180.)
def sind(x): return np.sin(x * np.pi / 180.)
def tand(x): return np.tan(x * np.pi / 180.)


var_g = {'sin': np.sin,
         'cos': np.cos,
         'tan': np.tan,
         'cosd': cosd,
         'sind': sind,
         'tand': tand,
         'arctan': np.arctan,
         'arctan2': np.arctan2,
         'exp': np.exp,
         'log10': np.log10,
         'log': np.log,
         'log2': np.log2,
         'sqrt': np.sqrt,
         'abs': np.abs,
         'conj': np.conj,
         'real': np.real,
         'imag': np.imag,
         'sum': np.sum,
         'dot': np.dot,
         'vdot': np.vdot,
         'array': np.array,
         'cross': np.cross,
         'pi': np.pi,
         'min': np.max,
         'min': np.min,
         'sign': np.sign,
         'ones': np.ones,
         'diag': np.diag,
         'zeros': np.zeros,
         'nan': np.nan,
         'inf': np.inf,
         'inv': np.linalg.inv,
         'linspace': np.linspace,
         'logspace': np.logspace}


def check_vectorfe_in_lowdim(gf):
    # check if this is VectorFE in lower dimensionality

    fes = gf.FESpace()
    sdim = fes.GetMesh().SpaceDimension()

    assert fes.GetNE() > 0, "Finite Element space has zero elements"

    isVector = (fes.GetFE(0).GetRangeType() == fes.GetFE(0).VECTOR)
    dim = fes.GetFE(0).GetDim()

    if isVector and dim < sdim:
        assert False, "Nodal evaluator does not work for low dimenstional vector field (try without averaging)"


class Variables(dict):
    def __repr__(self):
        txt = []
        for k in self.keys():
            txt.append(k + ':' + str(self[k]))
        return "\n".join(txt)


class Variable():
    '''
    define everything which we define algebra
    '''

    def __init__(self, complex=False, dependency=None, grad=None, curl=None, div=None):
        self.complex = complex

        # dependency stores a list of Finite Element space discrite variable
        # names whose set_point has to be called
        self.dependency = [] if dependency is None else dependency
        self.div = [] if div is None else div
        self.curl = [] if curl is None else curl
        self.grad = [] if grad is None else grad

    def __call__(self):
        raise NotImplementedError("Subclass need to implement")

    def __add__(self, other):
        if isinstance(other, Variable):
            return self() + other()
        return self() + other

    def __sub__(self, other):
        if isinstance(other, Variable):
            return self() - other()
        return self() - other

    def __mul__(self, other):
        if isinstance(other, Variable):
            return self() * other()
        return self() * other

    def __div__(self, other):
        if isinstance(other, Variable):
            return self() / other()
        return self() / other

    def __truediv__(self, other):
        if isinstance(other, Variable):
            return self() / other()
        return self() / other

    def __floordiv__(self, other):
        if isinstance(other, Variable):
            return self() // other()
        return self() // other

    def __radd__(self, other):
        if isinstance(other, Variable):
            return self() + other()
        return self() + other

    def __rsub__(self, other):
        if isinstance(other, Variable):
            return other() - self()
        return other - self()

    def __rmul__(self, other):
        if isinstance(other, Variable):
            return self() * other()
        return self() * other

    def __rdiv__(self, other):
        if isinstance(other, Variable):
            return other() / self()
        return other / self()

    def __rtruediv__(self, other):
        if isinstance(other, Variable):
            return other() / self()
        return other / self()

    def __rfloordiv__(self, other):
        if isinstance(other, Variable):
            return other() // self()
        return other // self()

    def __divmod__(self, other):
        if isinstance(other, Variable):
            return self().__divmod__(other())
        return self().__divmod__(other)

    def __floordiv__(self, other):
        if isinstance(other, Variable):
            return self().__floordiv__(other())
        return self().__floordiv__(other)

    def __mod__(self, other):
        if isinstance(other, Variable):
            return self().__mod__(other())
        return self().__mod__(other)

    def __pow__(self, other):
        if isinstance(other, Variable):
            return self().__pow__(other())
        return self().__pow__(other)

    def __neg__(self):
        return self().__neg__()

    def __pos__(self):
        return self().__pos__()

    def __abs__(self):
        return self().__abs__()

    def __getitem__(self, idx):
        return self()[idx]

    def get_names(self):
        return []

    def get_emesh_idx(self, idx=None, g=None):
        if idx is None:
            idx = []
        return idx

    def ncface_values(self, ifaces=None, irs=None,
                      gtypes=None, **kwargs):
        raise NotImplementedError("Subclass need to implement")

    def ncedge_values(self, *args, **kwargs):
        return self.ncface_values(*args, **kwargs)

    def point_values(self, *args, **kwargs):
        print(self)
        raise NotImplementedError("Subclass need to implement")

    def add_topological_info(self, mesh):
        if not isinstance(self, DomainVariable):
            return
        if mesh.Dimension() == 3:
            self.topo_info = (3, mesh.extended_connectivity['surf2vol'])
        if mesh.Dimension() == 2:
            self.topo_info = (2, mesh.extended_connectivity['line2surf'])
        if mesh.Dimension() == 1:
            self.topo_info = (1, mesh.extended_connectivity['vert2line'])

    def get_jitted_coefficient(self, *args):
        return None
    '''
    def make_callable(self):
        raise NotImplementedError("Subclass need to implement")

    def make_nodal(self):
        raise NotImplementedError("Subclass need to implement")
    '''


class TestVariable(Variable):
    def __init__(self, comp=-1, complex=False):
        super(TestVariable, self).__init__(complex=complex)

    def set_point(self, T, ip, g, l, t=None):
        self.x = T.Transform(ip)

    def __call__(self, **kwargs):
        return 2.

    def nodal_values(self, locs=None, **kwargs):
        # iele = None, elattr = None, el2v = None,
        #  wverts = None, locs = None, g = None
        return locs[:, 0] * 0 + 2.0

    def ncface_values(self, locs=None, **kwargs):
        return locs[:, 0] * 0 + 2.0


class Constant(Variable):
    def __init__(self, value, comp=-1):
        super(Constant, self).__init__(complex=np.iscomplexobj(value))
        self.value = value

    def __repr__(self):
        return "Constant(" + str(self.value) + ")"

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def set_point(self, T, ip, g, l, t=None):
        self.x = T.Transform(ip)

    def __call__(self, **kwargs):
        return self.value

    def nodal_values(self, iele=None, el2v=None, locs=None,
                     wverts=None, elvertloc=None, **kwargs):

        size = len(wverts)
        shape = [size] + list(np.array(self.value).shape)

        dtype = np.complex128 if self.complex else np.float64
        ret = np.zeros(shape, dtype=dtype)
        wverts = np.zeros(size)

        for kk, m, loc in zip(iele, el2v, elvertloc):
            if kk < 0:
                continue
            for pair, xyz in zip(m, loc):
                idx = pair[1]
                ret[idx] = self.value

        return ret

    def ncface_values(self, locs=None, **kwargs):
        size = len(locs)
        shape = [size] + list(np.array(self.value).shape)
        return np.tile(self.value, shape)

    def point_values(self, locs=None, **kwargs):
        size = len(locs)
        shape = [size] + list(np.array(self.value).shape)
        return np.tile(self.value, shape)


class SumVariable(Variable):
    def __init__(self, variables, gvariables):
        iscomplex = any([x.complex for x in variables])
        super(SumVariable, self).__init__(complex=iscomplex)
        self.variables = list(variables)
        self.gvariables = list(gvariables)

    def get_names(self):
        ret = []
        for x in self.variables:
            ret.extend(x.get_names())

        return ret

    def add_expression(self, v, g):
        self.complex = self.complex or v.complex
        self.variables.append(v)
        self.gvariables.append(g)

    def set_point(self, T, ip, _g, l, t=None):
        for v, g in zip(self.variables, self.gvariables):
            v.set_point(T, ip, g, l, t=t)

    def __call__(self, **kwargs):
        g = kwargs.pop("g", {}).copy()

        g.update(self.gvariables[0])
        v1 = self.variables[0].__call__(g=g, **kwargs)
        for v, g2 in zip(self.variables[1:], self.gvariables[1:]):
            g.update(g2)
            v1 = v1 + v.__call__(g=g, **kwargs)
        return v1

    def nodal_values(self, **kwargs):
        g = kwargs.pop("g", {}).copy()

        g.update(self.gvariables[0])
        v1 = self.variables[0].nodal_values(g=g, **kwargs)
        for v, g2 in zip(self.variables[1:], self.gvariables[1:]):
            g.update(g2)
            v1 = v1 + v.nodal_values(g=g, **kwargs)
        return v1

    def ncface_values(self, **kwargs):
        g = kwargs.pop("g", {}).copy()

        g.update(self.gvariables[0])
        v1 = self.variables[0].ncface_values(g=g, **kwargs)
        for v, g2 in zip(self.variables[1:], self.gvariables[1:]):
            g.update(g2)
            v1 = v1 + v.ncface_values(g=g, **kwargs)
        return v1

    def point_values(self, **kwargs):
        g = kwargs.pop("g", {}).copy()

        g.update(self.gvariables[0])
        v1 = self.variables[0].point_values(g=g, **kwargs)
        for v, g2 in zip(self.variables[1:], self.gvariables[1:]):
            g.update(g2)
            v1 = v1 + v.point_values(g=g, **kwargs)
        return v1


class CoordVariable(Variable):
    def __init__(self, comp=-1, complex=False):
        super(CoordVariable, self).__init__(complex=complex)
        self.comp = comp

    def __repr__(self):
        return "Coordinates"

    def set_point(self, T, ip, g, l, t=None):
        self.x = T.Transform(ip)

    def __call__(self, **kwargs):
        if self.comp == -1:
            return self.x
        else:
            return self.x[self.comp - 1]

    def nodal_values(self, locs=None, **kwargs):
        # iele = None, elattr = None, el2v = None,
        #  wverts = None, locs = None, g = None
        if self.comp == -1:
            return locs
        else:
            return locs[:, self.comp - 1]

    def ncface_values(self, locs=None, **kwargs):
        if self.comp == -1:
            return locs
        else:
            return locs[:, self.comp - 1]

    def point_values(self, counts=None, locs=None, **kwargs):
        if self.comp == -1:
            return locs
        else:
            return locs[:, self.comp - 1]


class ExpressionVariable(Variable):
    def __init__(self, expr, ind_vars, complex=False):
        super(ExpressionVariable, self).__init__(complex=complex)

        variables = []

        expr = expr.strip()

        code = compile(expr, '<string>', 'eval')
        names = code.co_names
        self.co = code
        self.names = names
        self.expr = expr
        self.ind_vars = ind_vars
        self.variables = WVD()
        # print 'Check Expression', expr.__repr__(), names

    def get_names(self):
        return self.names

    def __repr__(self):
        return "Expression(" + self.expr + ")"

    def set_point(self, T, ip, g, l, t=None):
        self.x = T.Transform(ip)
        #print("setting x", self, self.x)
        for n in self.names:
            if (n in g and isinstance(g[n], Variable)):
                g[n].set_point(T, ip, g, l, t=t)
                self.variables[n] = g[n]

    def __call__(self, **kwargs):
        l = {}
        for k, name in enumerate(self.ind_vars):
            l[name] = self.x[k]
        keys = self.variables.keys()
        for k in keys:
            l[k] = self.variables[k]()
        return (eval_code(self.co, var_g, l))

    def get_emesh_idx(self, idx=None, g=None):
        if idx is None:
            idx = []
        for n in self.names:
            if n in g and isinstance(g[n], Variable):
                idx = g[n].get_emesh_idx(idx=idx, g=g)
        return idx

    def nodal_values(self, iele=None, el2v=None, locs=None,
                     wverts=None, elvertloc=None, g=None,
                     **kwargs):
        #print("Entering nodal(expr)", self.expr)
        size = len(wverts)
        dtype = np.complex128 if self.complex else np.float64
        ret = np.zeros(size, dtype=dtype)
        for kk, m, loc in zip(iele, el2v, elvertloc):
            if kk < 0:
                continue
            for pair, xyz in zip(m, loc):
                idx = pair[1]
                ret[idx] = 1

        l = {}
        ll_name = []
        ll_value = []
        var_g2 = var_g.copy()

        for n in self.names:
            if (n in g and isinstance(g[n], Variable)):
                l[n] = g[n].nodal_values(iele=iele, el2v=el2v, locs=locs,
                                         wverts=wverts, elvertloc=elvertloc,
                                         g=g, **kwargs)
                ll_name.append(n)
                ll_value.append(l[n])
            elif (n in g):
                var_g2[n] = g[n]
        if len(ll_name) > 0:
            value = np.array([eval(self.co, var_g2, dict(zip(ll_name, v)))
                              for v in zip(*ll_value)])
        else:
            for k, name in enumerate(self.ind_vars):
                l[name] = locs[..., k]
            value = np.array(eval_code(self.co, var_g2, l), copy=False)
            if value.ndim > 1:
                value = np.stack([value] * size)
        #value = np.array(eval_code(self.co, var_g, l), copy=False)
        from petram.helper.right_broadcast import multi

        ret = multi(ret, value)
        #print("return (expr)", ret.shape)

        return ret

    def _ncx_values(self, method, ifaces=None, irs=None, gtypes=None,
                    g=None, attr1=None, attr2=None, locs=None,
                    **kwargs):

        size = len(locs)
        dtype = np.complex127 if self.complex else np.float64
        ret = np.zeros(size, dtype=dtype)

        l = {}
        ll_name = []
        ll_value = []
        var_g2 = var_g.copy()
        for n in self.names:
            if (n in g and isinstance(g[n], Variable)):
                m = getattr(g[n], method)
                # l[n] = g[n].ncface_values(ifaces = ifaces, irs = irs,
                l[n] = m(ifaces=ifaces, irs=irs,
                         gtypes=gtypes, locs=locs,
                         attr1=attr1, attr2=attr2,
                         g=g, **kwargs)
                ll_name.append(n)
                ll_value.append(l[n])
            elif (n in g):
                var_g2[n] = g[n]

        if len(ll_name) > 0:
            value = np.array([eval(self.co, var_g2, dict(zip(ll_name, v)))
                              for v in zip(*ll_value)])
        else:
            for k, name in enumerate(self.ind_vars):
                l[name] = locs[..., k]
            value = np.array(eval_code(self.co, var_g2, l), copy=False)
            if value.ndim > 1:
                value = np.stack([value] * size)

        return value

    def ncface_values(self, *args, **kwargs):
        return self._ncx_values('ncface_values', *args, **kwargs)

    def ncedge_values(self, *args, **kwargs):
        return self._ncx_values('ncedge_values', *args, **kwargs)

    def point_values(self, counts=None, locs=None, points=None,
                     attrs=None, elem_ids=None,
                     mesh=None, int_points=None, g=None,
                     knowns=None, **kwargs):

        l = {}
        ll_name = []
        ll_value = []
        var_g2 = var_g.copy()
        for n in self.names:
            if (n in g and isinstance(g[n], Variable)):
                l[n] = g[n].point_values(counts=counts, locs=locs, points=points,
                                         attrs=attrs, elem_ids=elem_ids,
                                         mesh=mesh, int_points=int_points, g=g,
                                         knowns=knowns)
                ll_name.append(n)
                ll_value.append(l[n])
            elif (n in g):
                var_g2[n] = g[n]
        if len(ll_name) > 0:
            value = np.array([eval(self.co, var_g2, dict(zip(ll_name, v)))
                              for v in zip(*ll_value)])
        else:
            for k, name in enumerate(self.ind_vars):
                l[name] = locs[..., k]
            value = np.array(eval_code(self.co, var_g2, l), copy=False)
            if value.ndim > 1:
                value = np.stack([value] * size)

        return value


class DomainVariable(Variable):
    def __init__(self, expr='', ind_vars=None, domains=None,
                 complex=False, gdomain=None):
        super(DomainVariable, self).__init__(complex=complex)
        self.domains = {}
        self.gdomains = {}
        if expr == '':
            return
        domains = sorted(domains)
        self.gdomains[tuple(domains)] = gdomain
        self.domains[tuple(domains)] = ExpressionVariable(expr, ind_vars,
                                                          complex=complex)

    def get_names(self):
        ret = []
        for x in self.domains:
            ret.extend(self.domains[x].get_names())
        return ret

    def __repr__(self):
        return "DomainVariable"

    def _add_something(self, something, gsomething, domains):
        doms = tuple(sorted(domains))

        existing_domains = list(self.domains)
        new_domains = tuple(np.setdiff1d(domains, sum(existing_domains, ())))

        new_exprs = {}
        new_gexprs = {}

        if len(new_domains) > 0:
            new_exprs[new_domains] = something
            new_gexprs[new_domains] = gsomething

        for ed in existing_domains:
            diff = tuple(np.setdiff1d(ed, doms))
            if len(diff) != 0:
                new_exprs[diff] = self.domains[ed]
                new_gexprs[diff] = self.gdomains[ed]
            insct = tuple(np.intersect1d(ed, doms))
            if len(insct) != 0:
                if isinstance(ed, SumVariable):
                    self.domains[ed].add_expression(something, gsomething)
                    new_exprs[insct] = self.domains[ed]
                    new_gexprs[insct] = {}
                else:
                    s = SumVariable(
                        (self.domains[ed], something), (self.gdomains[ed], gsomething))
                    new_exprs[insct] = s
                    new_gexprs[insct] = {}

        self.domains = new_exprs
        self.gdomains = new_gexprs

    def add_expression(self, expr, ind_vars, domains, gdomain, complex=False):
        new_expr = ExpressionVariable(expr, ind_vars,
                                      complex=complex)
        self._add_something(new_expr, gdomain, domains)
        if complex:
            self.complex = True

    def add_const(self, value, domains, gdomain):
        new_expr = Constant(value)
        self._add_something(new_expr, gdomain, domains)

        if np.iscomplexobj(value):
            self.complex = self.complex and True

    def set_point(self, T, ip, g, l, t=None):
        attr = T.Attribute
        if T.GetDimension() == self.topo_info[0]:
            # domain mode
            attrs = [attr]

        elif T.GetDimension() == self.topo_info[0] - 1:
            # boundary mode
            attrs = self.topo_info[1][attr]

        self.domain_target = []
        for domains in self.domains.keys():
            for a in attrs:
                if a in domains:
                    self.domains[domains].set_point(T, ip, g, l, t=t)
                self.domain_target.append(domains)

    def __call__(self, **kwargs):
        if len(self.domain_target) == 0:
            return 0.0
        # we return average for now. when domain variable is
        # evaluated on the boundary, it computes the aveage on both side.
        values = [self.domains[x]() for x in self.domain_target]
        ans = values[0]
        for v in values[1:]:
            ans = ans + v
        return ans / len(values)

    def get_emesh_idx(self, idx=None, g=None):
        if idx is None:
            idx = []
        for domains in self.domains.keys():
            expr = self.domains[domains]
            if isinstance(expr, Variable):
                gdomain = g if self.gdomains[domains] is None else self.gdomains[domains]
                idx.extend(expr.get_emesh_idx(idx=idx, g=gdomain))
        return idx

    def nodal_values(self, iele=None, elattr=None, g=None,
                     current_domain=None, **kwargs):
        # iele = None, elattr = None, el2v = None,
        # wverts = None, locs = None, g = None):
        from petram.helper.right_broadcast import add

        ret = None
        w = None

        for domains in self.domains.keys():
            if (current_domain is not None and
                    domains != current_domain):
                continue

            iele0 = np.zeros(iele.shape, dtype=int) - 1
            for domain in domains:
                idx = np.where(np.array(elattr) == domain)[0]
                iele0[idx] = iele[idx]

            expr = self.domains[domains]

            if self.gdomains[domains] is None:
                gdomain = g
            else:
                gdomain = g.copy()
                for key in self.gdomains[domains]:
                    gdomain[key] = self.gdomains[domains][key]

            v = expr.nodal_values(iele=iele0, elattr=elattr,
                                  current_domain=domains,
                                  g=gdomain, **kwargs)

            if w is None:
                a = np.sum(np.abs(v.reshape(len(v), -1)), -1)
                w = (a != 0).astype(float)
            else:
                a = np.sum(np.abs(v.reshape(len(v), -1)), -1)
                w = w + (a != 0).astype(float)

            #ret = v if ret is None else add(ret, v)
            ret = v if ret is None else ret + v

        idx = np.where(w != 0)[0]
        #ret2 = ret.copy()
        from petram.helper.right_broadcast import div

        ret[idx, ...] = div(ret[idx, ...], w[idx])

        return ret

    def _ncx_values(self, method, ifaces=None, irs=None, gtypes=None,
                    g=None, attr1=None, attr2=None, locs=None,
                    current_domain=None, **kwargs):
        from petram.helper.right_broadcast import add, multi

        ret = None

        w = ifaces * 0  # w : 0 , 0.5, 1
        for domains in self.domains:
            idx = np.in1d(attr1, domains)
            w[idx] = w[idx] + 1.0
            idx = np.in1d(attr2, domains)
            w[idx] = w[idx] + 1.0
        w[w > 0] = 1. / w[w > 0]

        npts = [irs[gtype].GetNPoints() for gtype in gtypes]
        base_weight = np.repeat(w, npts)
        # 1 for exterior face, 0.5 for internal faces

        for domains in self.domains.keys():
            if (current_domain is not None and
                    domains != current_domain):
                continue

            w = np.zeros(ifaces.shape)
            w[np.in1d(attr1, domains)] += 1.0
            w[np.in1d(attr2, domains)] += 1.0
            w2 = base_weight * np.repeat(w, npts)

            expr = self.domains[domains]

            if self.gdomains[domains] is None:
                gdomain = g
            else:
                gdomain = g.copy()
                gdomain.update(self.gdomains[domains])

            m = getattr(expr, method)
            #kwargs['weight'] = w2

            v = m(ifaces=ifaces, irs=irs,
                  gtypes=gtypes, locs=locs, attr1=attr1,
                  attr2=attr2, g=gdomain,
                  current_domain=domains,
                  **kwargs)

            v = multi(v, w2)
            ret = v if ret is None else ret + v
            #ret = v if ret is None else add(ret, v)
        return ret

    def ncface_values(self, *args, **kwargs):
        return self._ncx_values('ncface_values', *args, **kwargs)

    def ncedge_values(self, *args, **kwargs):
        return self._ncx_values('ncedge_values', *args, **kwargs)

    def point_values(self, counts=None, locs=None, points=None,
                     attrs=None, elem_ids=None,
                     mesh=None, int_points=None, g=None,
                     knowns=None, current_domain=None):

        valid_idx = np.where(attrs != -1)[0]
        valid_attrs = attrs[attrs != -1]
        ret = None

        for domains in self.domains.keys():
            if (current_domain is not None and
                    domains != current_domain):
                continue

            '''
            iele0 = np.zeros(iele.shape) - 1
            for domain in domains:
                idx = np.where(np.array(elattr) == domain)[0]
                iele0[idx] = iele[idx]
            '''
            expr = self.domains[domains]

            if self.gdomains[domains] is None:
                gdomain = g
            else:
                gdomain = g.copy()
                for key in self.gdomains[domains]:
                    gdomain[key] = self.gdomains[domains][key]

            idx = np.in1d(valid_attrs, domains)
            attrs2 = attrs.copy()
            attrs2[np.in1d(attrs, domains)] = -1
            v = expr.point_values(counts=counts, locs=locs, points=points,
                                  attrs=attrs, elem_ids=elem_ids,
                                  mesh=mesh, int_points=int_points, g=gdomain,
                                  current_domain=domains,
                                  knowns=knowns)
            if ret is None:
                ret = v
            else:
                idx = np.in1d(valid_attrs, domains)
                ret[idx] = v[idx]

        return ret


class PyFunctionVariable(Variable):
    def __init__(self, func, complex=False, shape=tuple(), dependency=None,
                 grad=None, curl=None, div=None):
        super(
            PyFunctionVariable,
            self).__init__(complex=complex,
                           dependency=dependency,
                           grad=grad,
                           curl=curl,
                           div=div)

        self.func = func
        self.t = None
        self.x = (0, 0, 0)
        self.shape = shape

    def __repr__(self):
        return "PyFunction"

    def set_point(self, T, ip, g, l, t=None):
        self.x = T.Transform(ip)
        self.t = t

    def __call__(self, **kwargs):
        if self.t is not None:
            args = tuple(np.hstack((self.x, t)))
        else:
            args = tuple(self.x)

        #kwargs = {n: locals()[n]() for n in self.dependency}
        # return np.array(self.func(*args, **kwargs), copy=False)
        return np.array(self.func(*args, **kwargs), copy=False)

    def nodal_values(self, iele=None, el2v=None, locs=None,
                     wverts=None, elvertloc=None, g=None, knowns=None,
                     **kwargs):
        # elattr = None, el2v = None,
        # wverts = None, locs = None, g = None
        if locs is None:
            return
        if g is None:
            g = {}
        if knowns is None:
            knowns = WKD()

        size = len(wverts)
        shape = [size] + list(self.shape)

        dtype = np.complex128 if self.complex else np.float64
        ret = np.zeros(shape, dtype=dtype)
        wverts = np.zeros(size)

        for kk, m, loc in zip(iele, el2v, elvertloc):
            if kk < 0:
                continue
            for pair, xyz in zip(m, loc):
                idx = pair[1]
                for n in self.dependency:
                    if not g[n] in knowns:
                        knowns[g[n]] = g[n].nodal_values(iele=iele, el2v=el2v,
                                                         locs=locs, wverts=wverts,
                                                         elvertloc=elvertloc,
                                                         g=g, knows=knowns,
                                                         **kwargs)
                kwargs = {n: knowns[g[n]][idx] for n in self.dependency}
                ret[idx] = ret[idx] + self.func(*xyz, **kwargs)
                wverts[idx] = wverts[idx] + 1
        ret = np.stack([x for x in ret if x is not None])

        idx = np.where(wverts == 0)[0]
        wverts[idx] = 1.0

        from petram.helper.right_broadcast import div
        ret = div(ret, wverts)

        return ret

    def _ncx_values(self, method, ifaces=None, irs=None, gtypes=None,
                    g=None, attr1=None, attr2=None, locs=None,
                    knowns=None, **kwargs):
        if locs is None:
            return
        if g is None:
            g = {}
        if knowns is None:
            knowns = WKD()

        dtype = np.complex128 if self.complex else np.float64

        ret = [None] * len(locs)
        for idx, xyz in enumerate(locs):
            '''
            for n in self.dependency:
                g[n].local_value = knowns[g[n]][idx]
                # putting the dependency variable to functions global.
                # this may not ideal, since there is potential danger
                # of name conflict?
                self.func.func_globals[n] = g[n]
            '''
            for n in self.dependency:
                if not g[n] in knowns:
                    m = getattr(g[n], method)
                    knowns[g[n]] = m(ifaces=ifaces, irs=irs,
                                     gtypes=gtypes, g=g,
                                     attr1=attr1, attr2=attr2,
                                     locs=locs, knows=knowns,
                                     **kwargs)

            kwargs = {n: knowns[g[n]][idx] for n in self.dependency}
            ret[idx] = self.func(*xyz, **kwargs)

        ret = np.stack(ret).astype(dtype, copy=False)

        return ret

    def ncface_values(self, *args, **kwargs):
        return self._ncx_values('ncface_values', *args, **kwargs)

    def ncedge_values(self, *args, **kwargs):
        return self._ncx_values('ncedge_values', *args, **kwargs)

    def point_values(self, counts=None, locs=None, points=None,
                     attrs=None, elem_ids=None,
                     mesh=None, int_points=None, g=None,
                     knowns=None, current_domains=None):

        if locs is None:
            return
        if g is None:
            g = {}
        if knowns is None:
            knowns = WKD()

        shape = [counts] + list(self.shape)
        dtype = np.complex128 if self.complex else np.float64
        ret = np.zeros(shape, dtype=dtype)

        valid_attrs = attrs[attrs != -1]

        jj = 0
        for i in range(counts):
            if valid_attrs[i] == -1:
                continue
            if (current_domains is not None and
                    not valid_attrs[i] in current_domains):
                continue

            for n in self.dependency:
                if not g[n] in knowns:
                    knowns[g[n]] = g[n].point_values(counts=counts, locs=locs, points=points,
                                                     attrs=attrs, elem_ids=elem_ids,
                                                     mesh=mesh, int_points=int_points, g=g,
                                                     knowns=knowns, current_domains=current_domains)

            xyz = tuple(locs[i])
            kwargs = {n: knowns[g[n]][i] for n in self.dependency}
            value = self.func(*xyz, **kwargs)

            ret[i, ...] = value

        return ret


class CoefficientVariable(Variable):
    def __init__(self, coeff_gen, l, g=None, coeff=None):
        self.coeff = coeff_gen(l, g)
        self.kind = coeff_gen.kind

        complex = not (self.coeff[1] is None)
        super(CoefficientVariable, self).__init__(complex=complex)

    @property
    def local_value(self):
        return self._local_value

    @local_value.setter
    def local_value(self, value):
        self._local_value = value

    def __call__(self, **kwargs):
        return self._local_value

    def set_point(self, T, ip, g, l, t=None):
        self.T = T
        self.ip = ip
        self.t = t
        self.set_local_from_T_ip()

    def set_local_from_T_ip(self):
        self.local_value = self.eval_local_from_T_ip()

    def eval_local_from_T_ip(self):
        call_eval = self.get_call_eval()

        T, ip = self.T, self.ip
        if (self.coeff[0] is not None and
                self.coeff[1] is not None):
            value = (np.array(call_eval(self.coeff[0], T, ip)) +
                     1j * np.array(self.coeff[1], T, ip))
        elif self.coeff[0] is not None:
            value = np.array(call_eval(self.coeff[0], T, ip))
        elif self.coeff[1] is not None:
            value = 1j * np.array(call_eval(self.coeff[1], T, ip))
        else:
            assert False, "coeff is (None, None)"
        return value

    def nodal_values(self, iele=None, ibele=None, elattr=None, el2v=None,
                     locs=None, elvertloc=None, wverts=None, mesh=None,
                     iverts_f=None, g=None, knowns=None,
                     edge_evaluator=False, **kwargs):

        g = mfem.Geometry()
        size = len(iverts_f)
        #wverts = np.zeros(size)
        ret = None
        if ibele is None:
            return

        if mesh.Dimension() == 3:
            if edge_evaluator:
                assert False, "EdgeNodal Evaluator does not supported dim=3"
            getelement = mesh.GetBdrElement
            gettransformation = mesh.GetBdrElementTransformation
        elif mesh.Dimension() == 2:
            if edge_evaluator:
                getelement = mesh.GetBdrElement
                gettransformation = mesh.GetBdrElementTransformation
            else:
                getelement = mesh.GetElement
                gettransformation = mesh.GetElementTransformation
        else:
            if edge_evaluator:
                getelement = mesh.GetElement
                gettransformation = mesh.GetElementTransformation
            else:
                assert False, "BdrNodal Evaluator does not support dim=1"

        call_eval = self.get_call_eval()

        for ibe in ibele:
            el = getelement(ibe)
            rule = g.GetVertices(el.GetGeometryType())
            nv = rule.GetNPoints()

            T = gettransformation(ibe)
            bverts = el.GetVerticesArray()

            for i in range(nv):
                ip = rule.IntPoint(i)
                T.SetIntPoint(ip)

                if (self.coeff[0] is not None and
                        self.coeff[1] is not None):
                    value = (np.array(call_eval(self.coeff[0], T, ip)) +
                             1j * np.array(call_eval(self.coeff[1], T, ip)))
                elif self.coeff[0] is not None:
                    value = np.array(call_eval(self.coeff[0], T, ip))
                elif self.coeff[1] is not None:
                    value = 1j * np.array(call_eval(self.coeff[1], T, ip))
                else:
                    assert False, "coeff is (None, None)"

                if ret is None:
                    if len(value.shape) == 0:
                        shape = np.hstack((size, ))
                    else:
                        shape = np.hstack((size, np.atleast_1d(value).shape))
                    ret = np.zeros(shape, dtype=value.dtype)

                idx = np.searchsorted(iverts_f, bverts[i])
                ret[idx, ...] = value

        return ret

    def ncface_values(self, ifaces=None, irs=None, gtypes=None, mesh=None,
                      **kwargs):

        call_eval = self.get_call_eval()

        if mesh.Dimension() == 3:
            m = mesh.GetFaceTransformation
        elif mesh.Dimension() == 2:
            m = mesh.GetElementTransformation

        data = []
        dtype = np.complex128 if self.complex else np.float64

        for i, gtype, in zip(ifaces, gtypes):
            T = m(i)
            ir = irs[gtype]
            nv = ir.GetNPoints()

            for j in range(nv):
                ip = ir.IntPoint(j)

                if (self.coeff[0] is not None and
                        self.coeff[1] is not None):
                    value = (np.array(call_eval(self.coeff[0], T, ip)) +
                             1j * np.array(call_eval(self.coeff[1], T, ip)))
                elif self.coeff[0] is not None:
                    value = np.array(call_eval(self.coeff[0], T, ip))
                elif self.coeff[1] is not None:
                    value = 1j * np.array(call_eval(self.coeff[1], T, ip))
                else:
                    assert False, "coeff is (None, None)"
                data.append(value)

        ret = np.stack(data).astype(dtype, copy=False)

        return ret

    def ncedge_values(self, ifaces=None, irs=None, gtypes=None, mesh=None,
                      **kwargs):

        call_eval = self.get_call_eval()

        if mesh.Dimension() == 2:
            m = mesh.GetFaceTransformation
        elif mesh.Dimension() == 1:
            m = mesh.GetElementTransformation
        else:
            assert False, "NCEdge Evaluator is not supported for this dimension"

        data = []
        dtype = np.complex128 if self.complex else np.float64

        for i, gtype, in zip(ifaces, gtypes):
            T = m(i)
            ir = irs[gtype]
            nv = ir.GetNPoints()

            for j in range(nv):
                ip = ir.IntPoint(j)

                if (self.coeff[0] is not None and
                        self.coeff[1] is not None):
                    value = (np.array(call_eval(self.coeff[0], T, ip)) +
                             1j * np.array(call_eval(self.coeff[1], T, ip)))
                elif self.coeff[0] is not None:
                    value = np.array(call_eval(self.coeff[0], T, ip))
                elif self.coeff[1] is not None:
                    value = 1j * np.array(call_eval(self.coeff[1], T, ip))
                else:
                    assert False, "coeff is (None, None)"
                data.append(value)

        ret = np.stack(data).astype(dtype, copy=False)

        return ret

    def point_values(self, counts=None, locs=None, points=None,
                     attrs=None, elem_ids=None,
                     mesh=None, int_points=None, g=None,
                     knowns=None, **kwargs):

        call_eval = self.get_call_eval()

        data = []
        dtype = np.complex128 if self.complex else np.float64

        for i in range(len(attrs)):
            if attrs[i] == -1:
                continue

            iele = elem_ids[i]
            T = mesh.GetElementTransformation(iele)
            ip = int_points[i]

            if (self.coeff[0] is not None and
                    self.coeff[1] is not None):
                value = (np.array(call_eval(self.coeff[0], T, ip)) +
                         1j * np.array(call_eval(self.coeff[1], T, ip)))
            elif self.coeff[0] is not None:
                value = np.array(call_eval(self.coeff[0], T, ip))
            elif self.coeff[1] is not None:
                value = 1j * np.array(call_eval(self.coeff[1], T, ip))
            else:
                assert False, "coeff is (None, None)"
            data.append(value)

        ret = np.stack(data).astype(dtype, copy=False)

        return ret

    def get_call_eval(self):
        if self.kind == "scalar":
            def call_eval(c, T, ip):
                return c.Eval(T, ip)
        elif self.kind == "vector":
            def call_eval(c, T, ip):
                v = mfem.Vector()
                c.Eval(v, T, ip)
                return v.GetDataArray().copy()
        elif self.kind == "matrix":
            def call_eval(c, T, ip):
                m = mfem.DenseMatrix()
                c.Eval(m, T, ip)
                return m.GetDataArray().copy()
        else:
            assert False, "unknown kind of Coefficient. Must be scalar/vector/matrix"
        return call_eval


class NumbaCoefficientVariable(CoefficientVariable):
    def __init__(self, func, complex=False, shape=tuple(), dependency=None,
                 grad=None, curl=None, div=None, td=False, params=None):
        super(
            CoefficientVariable,
            self).__init__(complex=complex,
                           dependency=dependency,
                           grad=grad,
                           curl=curl,
                           div=div)

        self.func = func
        self.params = params
        self.t = None
        self.x = (0, 0, 0)
        self.shape = shape
        self.td = td
        if len(self.shape) == 0:
            self.kind = 'scalar'
        elif len(self.shape) == 1:
            self.kind = 'vector'
        elif len(self.shape) == 2:
            self.kind = 'matrix'
        else:
            assert False, "unsupported shape"

        self._jitted = None

    def get_jitted_coefficient(self, ind_vars, locals):

        if self._jitted is not None:
            dprint1("(Note) this numba coefficient is already compiled", self.func)
            return self._jitted

        from petram.phys.numba_coefficient import NumbaCoefficient
        if isinstance(self.func, NumbaCoefficient):
            return self.func.mfem_numba_coeff

        from petram.helper.numba_utils import (generate_caller_scalar,
                                               generate_caller_array,
                                               generate_signature_scalar,
                                               generate_signature_array,)
        sdim = len(ind_vars)

        if len(self.shape) == 0:
            jitter = mfem.jit.scalar

            def gen_caller(setting):
                return generate_caller_scalar(setting, sdim)

            def gen_sig(setting):
                return generate_signature_scalar(setting, sdim)

            kwargs = {}

        elif len(self.shape) == 1:
            jitter = mfem.jit.vector

            def gen_caller(setting):
                return generate_caller_array(setting, sdim)

            def gen_sig(setting):
                return generate_signature_array(setting, sdim)

            kwargs = {"shape": self.shape}

        elif len(self.shape) == 2:
            jitter = mfem.jit.matrix

            def gen_caller(setting):
                return generate_caller_array(setting, sdim)

            def gen_sig(setting):
                return generate_signature_array(setting, sdim)

            kwargs = {"shape": self.shape}

        else:
            assert False, "unsupported shape"

        dep = []

        for d in self.dependency:
            dd = locals[d].get_jitted_coefficient(ind_vars, locals)
            if dd is None:
                return
            dep.append(dd)

        from petram.mfem_config import numba_debug, use_parallel
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank
        else:
            myid = 0
        numba_debug = False if myid != 0 else numba_debug

        wrapper = jitter(sdim=sdim,
                         complex=self.complex,
                         td=self.td,
                         params=self.params,
                         dependency=dep,
                         interface=(gen_caller, gen_sig),
                         debug=numba_debug,
                         **kwargs)
        self._jitted = wrapper(self.func)
        return self._jitted

    def set_coeff(self, ind_vars, locals):
        coeff = self.get_jitted_coefficient(ind_vars, locals)
        if self.complex:
            self.coeff = (coeff.real, coeff.imag)
        else:
            self.coeff = (coeff, None)


class GridFunctionVariable(Variable):
    def __init__(self, gf_real, gf_imag=None, comp=1,
                 deriv=None, complex=False):

        complex = not (gf_imag is None)
        super(GridFunctionVariable, self).__init__(complex=complex)
        self.dim = gf_real.VectorDim()
        self.comp = comp
        self.isGFSet = False
        self.isDerived = False
        self.deriv = deriv if deriv is not None else self._def_deriv
        self.deriv_args = (gf_real, gf_imag)

        self._curl = None
        self._div = None
        self._grad = None

    def _def_deriv(self, *args):
        return args[0], args[1], None

    def eval_grad(self):
        if self._grad is None:
            grad_r = mfem.GradientGridFunctionCoefficient(self.gfr)
            if self.gfi is not None:
                grad_i = mfem.GradientVectorGridFunctionCoefficient(self.gfi)
            else:
                grad_i = None
            self._grad_gf = (grad_r, grad_i)
        v = mfem.Vector()
        self._grad_gf[0].Eval(v, self.T, self.ip)
        ret = v.GetDataArray().copy()
        if self._grad_gf[1] is not None:
            self._grad_gf[1].Eval(v, self.T, self.ip)
            ret = ret + 1j*v.GetDataArray()
        return ret

    def eval_curl(self):
        if self._curl is None:
            curl_r = mfem.CurlGridFunctionCoefficient(self.gfr)
            if self.gfi is not None:
                curl_i = mfem.CurlVectorGridFunctionCoefficient(self.gfi)
            else:
                curl_i = None
            self._curl_gf = (curl_r, curl_i)
        v = mfem.Vector()
        self._curl_gf[0].Eval(v, self.T, self.ip)
        ret = v.GetDataArray().copy()
        if self._curl_gf[1] is not None:
            self._curl_gf[1].Eval(v, self.T, self.ip)
            ret = ret + 1j*v.GetDataArray()
        return ret

    def eval_div(self):
        if self._div is None:
            div_r = mfem.DivergenceGridFunctionCoefficient(self.gfr)
            if self.gfi is not None:
                div_i = mfem.DivergenceVectorGridFunctionCoefficient(self.gfi)
            else:
                div_i = None
            self._div_gf = (div_r, div_i)
        v = mfem.Vector()
        self._div_gf[0].Eval(v, self.T, self.ip)
        ret = v.GetDataArray().copy()
        if self._div_gf[1] is not None:
            self._div_gf[1].Eval(v, self.T, self.ip)
            ret = ret + 1j*v.GetDataArray()
        return ret

    def get_gf_real(self):
        if not self.isGFSet:
            self.set_gfr_gfi()
        return self.gfr

    def get_gf_imag(self):
        if not self.isGFSet:
            self.set_gfr_gfi()
        return self.gfi

    def set_gfr_gfi(self):
        gf_real, gf_imag, extra = self.deriv(*self.deriv_args)
        self.gfr = gf_real
        self.gfi = gf_imag
        self.extra = extra
        self.isGFSet = True
        return gf_real, gf_imag

    def set_point(self, T, ip, g, l, t=None):
        self.T = T
        self.ip = ip
        self.t = t
        self.set_local_from_T_ip()

    @property
    def local_value(self):
        return self._local_value

    @local_value.setter
    def local_value(self, value):
        self._local_value = value

    def __call__(self, **kwargs):
        return self._local_value

    def set_local_from_T_ip(self):
        self.local_value = self.eval_local_from_T_ip()

    def get_emesh_idx(self, idx=None, g=None):
        if idx is None:
            idx = []
        gf_real, gf_imag = self.deriv_args

        if gf_real is not None:
            if not gf_real._emesh_idx in idx:
                idx.append(gf_real._emesh_idx)
        elif gf_imag is not None:
            if not gf_imag._emesh_idx in idx:
                idx.append(gf_imag._emesh_idx)
        else:
            pass

        return idx

    def FESpace(self, check_parallel=True):
        gf_real, gf_imag = self.deriv_args
        if gf_real is not None:
            if hasattr(gf_real, "ParFESpace"):
                return gf_real.ParFESpace()
            else:
                return gf_real.FESpace()
        if gf_imag is not None:
            if hasattr(gf_imag, "ParFESpace"):
                return gf_imag.ParFESpace()
            else:
                return gf_imag.FESpace()


class GFScalarVariable(GridFunctionVariable):
    def __repr__(self):
        return "GridFunctionVariable (Scalar)"

    def set_funcs(self):
        # I should come back here to check if this works
        # with vector gf and/or boundary element. probably not...
        if not self.isGFSet:
            gf_real, gf_imag = self.set_gfr_gfi()
        else:
            gf_real, gf_imag = self.gfr, self.gfi

        name = gf_real.FESpace().FEColl().Name()
        if name.startswith("ND") or name.startswith("RT"):
            self.isVectorFE = True
            self.func_r = mfem.VectorGridFunctionCoefficient(gf_real)
            if gf_imag is not None:
                self.func_i = mfem.VectorGridFunctionCoefficient(gf_imag)
            else:
                self.func_i = None
        else:
            self.isVectorFE = False
            self.func_r = mfem.GridFunctionCoefficient(gf_real)
            if gf_imag is not None:
                self.func_i = mfem.GridFunctionCoefficient(gf_imag)
            else:
                self.func_i = None
        self.isDerived = True

    def eval_local_from_T_ip(self):
        if not self.isDerived:
            self.set_funcs()
        if self.isVectorFE:
            if self.func_i is None:
                v = mfem.Vector()
                self.func_r.Eval(v, self.T, self.ip)
                return v.GetDataArray()[self.comp - 1]
            else:
                v1 = mfem.Vector()
                v2 = mfem.Vector()
                self.func_r.Eval(v1, self.T, self.ip)
                self.func_i.Eval(v2, self.T, self.ip)
                return (v1.GetDataArray() + 1j *
                        v2.GetDataArray())[self.comp - 1]
        else:
            if self.func_i is None:
                return self.func_r.Eval(self.T, self.ip)
            else:
                return (self.func_r.Eval(self.T, self.ip) +
                        1j * self.func_i.Eval(self.T, self.ip))

    def nodal_values(self, iele=None, el2v=None, wverts=None,
                     **kwargs):
        if iele is None:
            return

        if not self.isDerived:
            self.set_funcs()

        # check if this is VectorFE in lower dimensionality
        gf = self.gfr if self.gfr is not None else self.gfi
        check_vectorfe_in_lowdim(gf)

        size = len(wverts)
        if self.gfi is None:
            ret = np.zeros(size, dtype=np.float64)
        else:
            ret = np.zeros(size, dtype=np.complex128)
        wverts = np.zeros(size)

        for kk, m in zip(iele, el2v):
            if kk < 0:
                continue
            values = mfem.doubleArray()

            self.gfr.GetNodalValues(kk, values, self.comp)

            for k, idx in m:
                ret[idx] = ret[idx] + values[k]
                wverts[idx] += 1
            if self.gfi is not None:
                arr = mfem.doubleArray()
                self.gfi.GetNodalValues(kk, arr, self.comp)
                for k, idx in m:
                    ret[idx] = ret[idx] + arr[k] * 1j

        ret = ret / wverts

        return ret

    def ncface_values(self, ifaces=None, irs=None,
                      gtypes=None, **kwargs):
        if not self.isDerived:
            self.set_funcs()

        name = self.gfr.FESpace().FEColl().Name()
        ndim = self.gfr.FESpace().GetMesh().Dimension()

        isVector = False
        if (name.startswith('RT') or
                name.startswith('ND')):
            d = mfem.DenseMatrix()
            p = mfem.DenseMatrix()
            isVector = True
        else:
            d = mfem.Vector()
            p = mfem.DenseMatrix()
        data = []

        def get_method(gf, ndim, isVector):
            if gf is None:
                return None
            if ndim == 3:
                if isVector:
                    return gf.GetFaceVectorValues
                elif gf.VectorDim() > 1:
                    def func(i, side, ir, vals, tr, in_gf=gf):
                        in_gf.GetFaceValues(
                            i, side, ir, vals, tr, vdim=self.comp)
                    return func
                else:
                    return gf.GetFaceValues
            elif ndim == 2:
                if isVector:
                    def func(i, side, ir, vals, tr, in_gf=gf):
                        in_gf.GetVectorValues(i, ir, vals, tr)
                    return func
                elif gf.VectorDim() > 1:
                    def func(i, side, ir, vals, tr, in_gf=gf):
                        #in_gf.GetValues(i, ir, vals, tr, vdim=self.comp - 1)
                        in_gf.GetValues(i, ir, vals, tr, self.comp-1)
                    return func
                else:
                    def func(i, side, ir, vals, tr, in_gf=gf):
                        in_gf.GetValues(i, ir, vals, tr)
                        return
                    return func
            else:
                assert False, "ndim = 1 has no face"
            return None

        getvalr = get_method(self.gfr, ndim, isVector)
        getvali = get_method(self.gfi, ndim, isVector)

        for i, gtype, in zip(ifaces, gtypes):
            ir = irs[gtype]
            getvalr(i, 2, ir, d, p)  # side = 2 (automatic?)
            v = d.GetDataArray().copy()
            if isVector:
                v = v[self.comp - 1, :]

            if getvali is not None:
                getvali(i, 2, ir, d, p)  # side = 2 (automatic?)
                vi = d.GetDataArray().copy()
                if isVector:
                    vi = vi[self.comp - 1, :]
                v = v + 1j * vi
            data.append(v)
        data = np.hstack(data)

        return data

    def ncedge_values(self, ifaces=None, irs=None,
                      gtypes=None, **kwargs):
        if not self.isDerived:
            self.set_funcs()

        name = self.gfr.FESpace().FEColl().Name()
        ndim = self.gfr.FESpace().GetMesh().Dimension()

        isVector = False
        if name.startswith('RT'):
            d = mfem.DenseMatrix()
            p = mfem.DenseMatrix()
            isVector = True
        elif name.startswith('ND'):
            d = mfem.DenseMatrix()
            p = mfem.DenseMatrix()
            isVector = True
        else:
            d = mfem.Vector()
            p = mfem.DenseMatrix()
        data = []

        def get_method(gf, ndim, isVector):
            if gf is None:
                return None
            if ndim == 1:
                # if isVector:
                #    def func(i, ir, vals, tr, in_gf=gf):
                if gf.VectorDim() > 1:
                    def func(i, ir, vals, tr, in_gf=gf):
                        in_gf.GetValues(i, ir, vals, tr, self.comp - 1)
                    return func
                else:
                    def func(i, ir, vals, tr, in_gf=gf):
                        in_gf.GetValues(i, ir, vals, tr)
                        return
                    return func
            elif ndim == 2:
                side = 2
                if isVector:
                    def func(i, ir, vals, tr, in_gf=gf):
                        in_gf.GetFaceVectorValues(i, side, ir, vals, tr)
                    return func
                elif gf.VectorDim() > 1:
                    def func(i, ir, vals, tr, in_gf=gf):
                        in_gf.GetFaceValues(
                            i, side, ir, vals, tr, self.comp - 1)
                    return func
                else:
                    def func(i, ir, vals, tr, in_gf=gf):
                        in_gf.GetFaceValues(i, side, ir, vals, tr)
                        return
                    return func

            else:
                assert False, "ndim = 3 is not supported"
            return None

        getvalr = get_method(self.gfr, ndim, isVector)
        getvali = get_method(self.gfi, ndim, isVector)

        for i, gtype, in zip(ifaces, gtypes):
            ir = irs[gtype]
            getvalr(i, ir, d, p)  # side = 2 (automatic?)
            v = d.GetDataArray().copy()
            if isVector:
                v = v[self.comp - 1, :]

            if getvali is not None:
                getvali(i, ir, d, p)  # side = 2 (automatic?)
                vi = d.GetDataArray().copy()
                if isVector:
                    vi = vi[self.comp - 1, :]
                v = v + 1j * vi
            data.append(v)

        data = np.hstack(data)
        return data

    def point_values(self, counts=None, locs=None, points=None,
                     attrs=None, elem_ids=None,
                     mesh=None, int_points=None, g=None,
                     knowns=None, **kwargs):

        if not self.isDerived:
            self.set_funcs()
        gf = self.gfr if self.gfr is not None else self.gfi

        name = gf.FESpace().FEColl().Name()
        ndim = gf.FESpace().GetMesh().Dimension()

        if (name.startswith('RT') or
                name.startswith('ND')):
            isVector = True
        else:
            isVector = False

        d = mfem.Vector()

        if self.complex:
            dtype = complex
            val = complex(0.0)
        else:
            dtype = float
            val = float(0.0)

        data = np.zeros(counts, dtype=dtype)

        jj = 0
        for i in range(len(attrs)):
            if attrs[i] == -1:
                continue

            ip = int_points[i]
            iele = elem_ids[i]

            val = val * 0.

            if self.gfr is not None:
                if isVector:
                    self.gfr.GetVectorValue(iele, ip, d)
                    val = d[self.comp - 1]
                else:
                    val = self.gfr.GetValue(iele, ip, self.comp - 1)

            if self.gfi is not None:
                if isVector:
                    self.gfi.GetVectorValue(iele, ip, d)
                    val += 1j * d[self.comp - 1]
                else:
                    val += 1j * self.gfi.GetValue(iele, ip, self.comp - 1)

            data[jj] = val
            jj = jj + 1

        return data

    def get_jitted_coefficient(self, ind_vars, locals):
        if not self.isDerived:
            self.set_funcs()

        if isinstance(self.func_r, mfem.VectorCoefficient):
            v = [0] * self.func_r.GetVDim()
            v[self.comp-1] = 1
            c2 = mfem.VectorConstantCoefficient(v)
            # the value of c2 will be copied.
            ret1 = mfem.InnerProductCoefficient(self.func_r, c2)
            if self.func_i is not None:
                ret2 = mfem.InnerProductCoefficient(self.func_i, c2)
                return (ret1, ret2)
            else:
                return ret1

        if self.func_i is None:
            return self.func_r
        else:
            return (self.func_r, self.func_i)


class GFVectorVariable(GridFunctionVariable):
    def __repr__(self):
        return "GridFunctionVariable (Vector)"

    def set_funcs(self):
        if not self.isGFSet:
            gf_real, gf_imag = self.set_gfr_gfi()
        else:
            gf_real, gf_imag = self.gfr, self.gfi

        self.dim = gf_real.VectorDim()
        name = gf_real.FESpace().FEColl().Name()
        if name.startswith("ND") or name.startswith("RT"):
            self.isVectorFE = True
            self.func_r = mfem.VectorGridFunctionCoefficient(gf_real)
            if gf_imag is not None:
                self.func_i = mfem.VectorGridFunctionCoefficient(gf_imag)
            else:
                self.func_i = None

        else:
            self.isVectorFE = False
            self.func_r = [mfem.GridFunctionCoefficient(gf_real, comp=k + 1)
                           for k in range(self.dim)]

            if gf_imag is not None:
                self.func_i = [mfem.GridFunctionCoefficient(gf_imag, comp=k + 1)
                               for k in range(self.dim)]
            else:
                self.func_i = None
        self.isDerived = True

    def eval_local_from_T_ip(self):
        if not self.isDerived:
            self.set_funcs()
        if self.isVectorFE:
            if self.func_i is None:
                v = mfem.Vector()
                self.func_r.Eval(v, self.T, self.ip)
                return v.GetDataArray().copy()
            else:
                v1 = mfem.Vector()
                v2 = mfem.Vector()
                self.func_r.Eval(v1, self.T, self.ip)
                self.func_i.Eval(v2, self.T, self.ip)
                return v1.GetDataArray().copy() + 1j * v2.GetDataArray().copy()
        else:
            if self.func_i is None:
                return np.array([func_r.Eval(self.T, self.ip) for
                                 func_r in self.func_r])
            else:
                return np.array([(func_r.Eval(self.T, self.ip) +
                                  1j * func_i.Eval(self.T, self.ip))
                                 for func_r, func_i
                                 in zip(self.func_r, self.func_i)])

    def nodal_values(self, iele=None, el2v=None, wverts=None,
                     **kwargs):
        # iele = None, elattr = None, el2v = None,
        # wverts = None, locs = None, g = None
        #print("grid vector nodal")
        if iele is None:
            return
        if not self.isDerived:
            self.set_funcs()

        size = len(wverts)

        # check if this is VectorFE in lower dimensionality
        gf = self.gfr if self.gfr is not None else self.gfi
        check_vectorfe_in_lowdim(gf)

        ans = []
        for comp in range(self.dim):
            if self.gfi is None:
                ret = np.zeros(size, dtype=np.float64)
            else:
                ret = np.zeros(size, dtype=np.complex128)

            wverts = np.zeros(size)
            for kk, m in zip(iele, el2v):
                if kk < 0:
                    continue
                values = mfem.doubleArray()
                self.gfr.GetNodalValues(kk, values, comp + 1)
                for k, idx in m:
                    ret[idx] = ret[idx] + values[k]
                    wverts[idx] += 1
                if self.gfi is not None:
                    arr = mfem.doubleArray()
                    self.gfi.GetNodalValues(kk, arr, comp + 1)
                    for k, idx in m:
                        ret[idx] = ret[idx] + arr[k] * 1j
            # print(list(wverts))
            ans.append(ret / wverts)
        ret = np.transpose(np.vstack(ans))
        return ret

    def ncface_values(self, ifaces=None, irs=None,
                      gtypes=None, **kwargs):

        if not self.isDerived:
            self.set_funcs()
        ndim = self.gfr.FESpace().GetMesh().Dimension()

        d = mfem.DenseMatrix()
        p = mfem.DenseMatrix()
        data = []

        def get_method(gf, ndim):
            if gf is None:
                return None
            if ndim == 3:
                return gf.GetFaceVectorValues
            elif ndim == 2:
                return gf.GetVectorValues
            else:
                assert False, "ndim = 1 has no face"
        getvalr = get_method(self.gfr, ndim)
        getvali = get_method(self.gfi, ndim)

        for i, gtype, in zip(ifaces, gtypes):
            ir = irs[gtype]
            getvalr(i, 2, ir, d, p)  # side = 2 (automatic?)
            v = d.GetDataArray().copy()

            if getvali is not None:
                getvali(i, 2, ir, d, p)
                vi = d.GetDataArray().copy()
                v = v + 1j * vi
            data.append(v)
        ret = np.hstack(data).transpose()

        return ret

    def point_values(self, counts=None, locs=None, points=None,
                     attrs=None, elem_ids=None,
                     mesh=None, int_points=None, g=None,
                     knowns=None, **kwargs):

        if not self.isDerived:
            self.set_funcs()
        gf = self.gfr if self.gfr is not None else self.gfi

        name = gf.FESpace().FEColl().Name()
        ndim = gf.FESpace().GetMesh().Dimension()

        if (name.startswith('RT') or
                name.startswith('ND')):
            isVector = True
            vdim = gf.FESpace().GetMesh().SpaceDimension()
        else:
            isVector = False
            vdim = gf.VectorDim()

        d = mfem.Vector()

        if self.complex:
            dtype = complex

        else:
            dtype = float

        data = np.zeros((counts, vdim), dtype=dtype)

        jj = 0
        for i in range(len(attrs)):
            if attrs[i] == -1:
                continue

            ip = int_points[i]
            iele = elem_ids[i]

            val = 0.0

            if self.gfr is not None:
                self.gfr.GetVectorValue(iele, ip, d)
                val = d.GetDataArray().copy()

            if self.gfi is not None:
                self.gfi.GetVectorValue(iele, ip, d)
                dd = 1j * d.GetDataArray().copy()
                val = val + dd

            data[jj, :] = val
            jj = jj + 1

        return data


'''

Surf Variable:
 Regular Variable + Surface Geometry (n, nx, ny, nz)

'''


class SurfVariable(Variable):
    def __init__(self, sdim, complex=False):
        self.sdim = sdim
        super(SurfVariable, self).__init__(complex=complex)


class SurfNormal(SurfVariable):
    def __init__(self, sdim, comp=-1, complex=False):
        self.comp = comp
        SurfVariable.__init__(self, sdim, complex=complex)

    def __repr__(self):
        return "SurfaceNormal (nx, ny, nz)"

    def set_point(self, T, ip, g, l, t=None):
        nor = mfem.Vector(self.sdim)
        mfem.CalcOrtho(T.Jacobian(), nor)
        self.nor = nor.GetDataArray().copy()

    def get_jitted_coefficient(self, ind_vars, locals):
        norm = mfem.VectorBdrNormalCoefficient(len(ind_vars))
        return norm

    def __call__(self, **kwargs):
        if self.comp == -1:
            return self.nor
        else:
            return self.nor[self.comp - 1]

    def nodal_values(self, ibele=None, mesh=None, iverts_f=None,
                     **kwargs):
        # iele = None, elattr = None, el2v = None,
        # wverts = None, locs = None, g = None

        g = mfem.Geometry()
        size = len(iverts_f)
        #wverts = np.zeros(size)
        ret = np.zeros((size, self.sdim))
        if ibele is None:
            return

        for ibe in ibele:
            el = mesh.GetBdrElement(ibe)
            rule = g.GetVertices(el.GetGeometryType())
            nv = rule.GetNPoints()

            T = mesh.GetBdrElementTransformation(ibe)
            bverts = mesh.GetBdrElement(ibe).GetVerticesArray()

            for i in range(nv):
                nor = mfem.Vector(self.sdim)
                T.SetIntPoint(rule.IntPoint(i))
                mfem.CalcOrtho(T.Jacobian(), nor)
                idx = np.searchsorted(iverts_f, bverts[i])

                ret[idx, :] += nor.GetDataArray().copy()
                #wverts[idx] = wverts[idx] + 1

        #for i in range(self.sdim): ret[:,i] /= wvert
        # normalize to length one.
        ret = ret / np.sqrt(np.sum(ret**2, 1)).reshape(-1, 1)

        if self.comp == -1:
            return ret
        return ret[:, self.comp - 1]

    def ncface_values(self, ifaces=None, irs=None, gtypes=None,
                      locs=None, mesh=None, **kwargs):

        size = len(locs)
        ret = np.zeros((size, self.sdim))
        if ifaces is None:
            return

        nor = mfem.Vector(self.sdim)

        if mesh.Dimension() == 3:
            m = mesh.GetFaceTransformation
        elif mesh.Dimension() == 2:
            m = mesh.GetElementTransformation
        idx = 0
        for i, gtype, in zip(ifaces, gtypes):
            ir = irs[gtype]
            nv = ir.GetNPoints()
            T = m(i)
            for j in range(nv):
                T.SetIntPoint(ir.IntPoint(i))
                mfem.CalcOrtho(T.Jacobian(), nor)
                ret[idx, :] = nor.GetDataArray().copy()
                idx = idx + 1

        from petram.helper.right_broadcast import div

        ret = div(ret, np.sqrt(np.sum(ret**2, -1)))
        if self.comp == -1:
            return ret
        return ret[:, self.comp - 1]

    def ncedge_values(self, *args, **kwargs):
        raise NotImplementedError("Normal is not defined on Edge")


class SurfExpressionVariable(ExpressionVariable, SurfVariable):
    '''
    expression valid on surface
    '''

    def __init__(self, expr, ind_vars, sdim, complex=False):
        ExpressionVariable.__init__(self, expr, ind_vars, complex=complex)
        SurfVariable.__init__(self, sdim, complex=complex)

    def __repr__(self):
        return "SurfaceExpression(" + self.expr + ")"

    def set_point(self, T, ip, g, l, t=None):
        self.x = T.Transform(ip)
        self.t = t
        T.SetIntPoint(ip)
        nor = mfem.Vector(self.sdim)
        mfem.CalcOrtho(T.Jacobian(), nor)
        self.nor = nor.GetDataArray().copy()

    def __call__(self, **kwargs):
        l = {}
        for k, name in enumerate(self.ind_vars):
            l[name] = self.x[k]
        l['n'] = self.nor
        for k, name in enumerate(self.ind_vars):
            l['n' + name] = self.nor[k]
        keys = self.variables.keys()
        for k in keys:
            l[k] = self.variables[k]()
        return (eval_code(self.co, var_g, l))

    def nodal_values(self, **kwargs):
        # this may not be used al all??

        l = {}
        for n in self.names:
            if (n in g and isinstance(g[n], Variable)):
                l[n] = g[n].nodal_values(**kwargs)
        for k, name in enumerate(self.ind_vars):
            l[name] = locs[..., k]
        for k, name in enumerate(self.ind_vars):
            l['n' + name] = nor[..., k]
        return (eval_code(self.co, var_g, l))

    def ncface_values(self, **kwargs):
        assert False, "ncface in SurfaceExpressionVariable must be added"


'''
 Bdr Variable = Surface Variable defined on particular boundary
'''


class BdrVariable(ExpressionVariable, SurfVariable):
    pass


def append_suffix_to_expression(expr, vars, suffix):
    if vars is None:
        return expr + suffix
    for v in vars:
        expr = expr.replace(v, v + suffix)
    return expr


def add_scalar(solvar, name, suffix, ind_vars, solr,
               soli=None, deriv=None, vars=None):
    name = append_suffix_to_expression(name, vars, suffix)
    solvar[name] = GFScalarVariable(solr, soli, comp=1, deriv=deriv)


def add_components(solvar, name, suffix, ind_vars, solr,
                   soli=None, deriv=None, vars=None):
    name = append_suffix_to_expression(name, vars, suffix)
    solvar[name] = GFVectorVariable(solr, soli, deriv=deriv)
    for k, p in enumerate(ind_vars):
        solvar[name + p] = GFScalarVariable(solr, soli, comp=k + 1,
                                            deriv=deriv)


def add_elements(solvar, name, suffix, ind_vars, solr,
                 soli=None, deriv=None, elements=None):
    elements = elements if elements is not None else []
    for k, p in enumerate(ind_vars):
        solvar[name + suffix + p] = GFScalarVariable(solr, soli, comp=k + 1,
                                                     deriv=deriv)


def add_component_expression(solvar, name, suffix, ind_vars, expr, vars,
                             componentname,
                             domains=None, bdrs=None, complex=None,
                             gdomain=None, gbdr=None):
    expr = append_suffix_to_expression(expr, vars, suffix)

    if isinstance(componentname, int):
        componentname = ind_vars[componentname]

    cname = name + suffix + componentname
    if domains is not None:
        if (cname) in solvar:
            solvar[cname].add_expression(expr, ind_vars, domains,
                                         gdomain,
                                         complex=complex)
        else:
            solvar[cname] = DomainVariable(expr, ind_vars,
                                           domains=domains,
                                           complex=complex,
                                           gdomain=gdomain)
    elif bdrs is not None:
        assert False, "BoundaryVariable not implemented."

    else:
        solvar[cname] = ExpressionVariable(expr, ind_vars,
                                           complex=complex)


def add_expression(solvar, name, suffix, ind_vars, expr, vars,
                   domains=None, bdrs=None, complex=None,
                   gdomain=None, gbdr=None):

    expr = append_suffix_to_expression(expr, vars, suffix)

    if domains is not None:
        if (name + suffix) in solvar:
            solvar[name + suffix].add_expression(expr, ind_vars, domains,
                                                 gdomain,
                                                 complex=complex)

        else:
            solvar[name + suffix] = DomainVariable(expr, ind_vars,
                                                   domains=domains,
                                                   complex=complex,
                                                   gdomain=gdomain)
    elif bdrs is not None:
        assert False, "BoundaryVariable not implemented."

    else:
        solvar[name + suffix] = ExpressionVariable(expr, ind_vars,
                                                   complex=complex)


def add_constant(solvar, name, suffix, value, domains=None,
                 gdomain=None, bdrs=None, gbdr=None):

    if domains is not None:
        if (name + suffix) in solvar:
            solvar[name + suffix].add_const(value, domains, gdomain)
        else:
            solvar[name + suffix] = DomainVariable('')
            solvar[name + suffix].add_const(value, domains, gdomain)
    elif bdrs is not None:
        pass
    else:
        solvar[name + suffix] = Constant(value)
        #solvar[name + suffix] = value


def add_surf_normals(solvar, ind_vars):
    sdim = len(ind_vars)
    solvar['n'] = SurfNormal(sdim, comp=-1)
    for k, p in enumerate(ind_vars):
        solvar['n' + p] = SurfNormal(sdim, comp=k + 1)


def add_coordinates(solvar, ind_vars):
    for k, p in enumerate(ind_vars):
        solvar[p] = CoordVariable(comp=k + 1)


def project_variable_to_gf(c, ind_vars, gfr, gfi,
                           global_ns=None, local_ns=None):

    if global_ns is None:
        global_ns = {}
    if local_ns is None:
        local_ns = {}

    from petram.phys.weakform import VCoeff, SCoeff

    fes = gfr.FESpace()
    ndim = fes.GetMesh().Dimension()
    sdim = fes.GetMesh().SpaceDimension()
    vdim = fes.GetVDim()
    fec = fes.FEColl().Name()

    if (fec.startswith('ND') or fec.startswith('RT')):
        coeff_dim = sdim
    else:
        coeff_dim = vdim

    def project_coeff(gf, coeff_dim, c, ind_vars, real):
        if coeff_dim > 1:
            #print("vector coeff", c)
            coeff = VCoeff(coeff_dim, c, ind_vars,
                           local_ns, global_ns, real=real)
        else:
            #print("coeff", c)
            coeff = SCoeff(c, ind_vars,
                           local_ns, global_ns, real=real)

        gf.ProjectCoefficient(coeff)

    project_coeff(gfr, coeff_dim, c, ind_vars, True)
    if gfi is not None:
        project_coeff(gfi, coeff_dim, c, ind_vars, False)


'''

   NativeCoefficient class

   This class opens the possibility ot use mfem native coefficient (C++)
   class object in BF/LF.

   We can define  math operatios between native coefficent class objects
   in the way to map the operatio to SumCoefficient/ProductCofficient/...
   recently added in MFEM.

   Full implementation needs to wait update of PyMFEM. Eventually, this
   class may move to PyMFEM (under mfem.common)

   from petram.helper.variables import variable, coefficient
   @coefficient.complex()
   def ksqrt():
       coeff1 = mfem.ConstantCoefficient(900)
       coeff2 = mfem.ConstantCoefficient(200)
       return coeff1, coeff2
'''


class _coeff_decorator(object):
    def float(self, dependency=None, td=False, jit=False):
        def dec(func):
            obj = NativeCoefficientGen(
                func, dependency=dependency, td=td, jit=jit)
            return obj
        return dec

    def complex(self, dependency=None, td=False, jit=False):
        def dec(func):
            obj = ComplexNativeCoefficientGen(
                func, dependency=dependency, td=td, jit=jit)
            return obj
        return dec

    def array(self, complex=False, shape=(1,),
              dependency=None, td=False, jit=False):
        def dec(func):
            if len(shape) == 1:
                if complex:
                    obj = VectorComplexNativeCoefficientGen(
                        func, dependency=dependency, shape=shape, jit=jit, td=td)
                else:
                    obj = VectorNativeCoefficientGen(
                        func, dependency=dependency, shape=shape, jit=jit, td=td)
            elif len(shape) == 2:
                if complex:
                    obj = MatrixComplexNativeCoefficientGen(
                        func, dependency=dependency, shape=shape, jit=jit, td=td)
                else:
                    obj = MatrixNativeCoefficientGen(
                        func, dependency=dependency, shape=shape, jit=jit, td=td)
            return obj
        return dec


coefficient = _coeff_decorator()


class NativeCoefficientGenBase(object):
    '''
    define everything which we define algebra
    '''

    def __init__(self, fgen, igen=None, complex=False,
                 dependency=None, shape=None, td=False, jit=False):
        self.complex = complex
        # dependency stores a list of Finite Element space discrite variable
        # names whose set_point has to be called
        self.dependency = [] if dependency is None else dependency
        self.fgen = fgen
        self.shape = shape
        self.complex = complex
        self.jit = jit
        self.td = False
        self._generated = None

    def __call__(self, l, g=None):
        '''
        call fgen to generate coefficient

        '''
        if self._generated is not None:
            return self._generated
        m = getattr(self, 'fgen')
        if not self.jit:
            args = []
            for n in self.dependency:
                if self.complex:
                    args.append((l[n].get_gf_real(), l[n].get_gf_imag()))
                else:
                    args.append(l[n].get_gf_real())

            rc = m(*args)
            if self.complex:
                if len(rc) != 2:
                    assert False, "generator must return real/imag parts"
                self._generated = rc
                return self._generated
            else:
                self._generated = rc, None
                return self._generated
        else:
            assert False, "coefficient(jit=True) is not valid anymore. jit is supported by @variable"

    def scale_coeff(self, coeff, scale):
        if self.shape is None:
            c2 = mfem.ConstantCoefficient(scale)
            ret = mfem.ProductCoefficient(coeff, c2)
            ret._c2 = c2
            ret._coeff = coeff
            return ret
        elif len(self.shape) == 1:  # Vector
            c2 = mfem.ConstantCoefficient(scale)
            ret = mfem.ScalarVectorProductCoefficient(c2, coeff)
            ret._c2 = c2
            ret._coeff = coeff
            return ret
        elif len(self.shape) == 2:  # Matrix
            c2 = mfem.ConstantCoefficient(scale)
            ret = mfem.ScalarMatrixProductCoefficient(c2, coeff)
            ret._c2 = c2
            ret._coeff = coeff
            return ret
        else:
            assert False, "dim >= 3 is not supported"

    def get_emesh_idx(self, idx=None, g=None):
        if idx is None:
            idx = []
        return idx


class NativeCoefficientGen(NativeCoefficientGenBase):
    kind = "scalar"

    def __init__(self, func, dependency=None, td=False, jit=False):
        NativeCoefficientGenBase.__init__(
            self, func, complex=False, dependency=dependency, td=td, jit=jit)


class ComplexNativeCoefficientGen(NativeCoefficientGenBase):
    kind = "scalar"

    def __init__(self, func, dependency=None, td=False, jit=False):
        NativeCoefficientGenBase.__init__(
            self, func, complex=True, dependency=dependency, td=td, jit=jit)


class VectorNativeCoefficientGen(NativeCoefficientGenBase):
    kind = "vector"

    def __init__(self, func, dependency=None, shape=None, td=False, jit=False):
        NativeCoefficientGenBase.__init__(
            self, func, complex=False, dependency=dependency, shape=shape, td=td, jit=jit)


class VectorComplexNativeCoefficientGen(NativeCoefficientGenBase):
    kind = "vector"

    def __init__(self, func, dependency=None, shape=None, td=False, jit=False):
        NativeCoefficientGenBase.__init__(
            self, func, complex=True, dependency=dependency, shape=shape, td=td, jit=jit)


class MatrixNativeCoefficientGen(NativeCoefficientGenBase):
    kind = "matrix"

    def __init__(self, func, dependency=None, shape=None, td=False, jit=False):
        NativeCoefficientGenBase.__init__(
            self, func, complex=False, dependency=dependency, shape=shape, td=td, jit=jit)


class MatrixComplexNativeCoefficientGen(NativeCoefficientGenBase):
    kind = "matrix"

    def __init__(self, func, dependency=None, shape=None, td=False, jit=False):
        NativeCoefficientGenBase.__init__(
            self, func, complex=True, dependency=dependency, shape=shape, td=td, jit=jit)



model = proj.model1.mfem.param.getvar("mfem_model")
for m in model.walk():
    mname = m.__class__.__module__
    tmp = mname.split('.'); tmp[0] = 'petram'
    mname = '.'.join(tmp)
    mod = __import__(mname, globals(), locals(), [m.__class__.__name__])

    c = getattr(mod, m.__class__.__name__)
    #print m.__class__, c
    m.__class__ = c

#model.set_guiscript('.scripts.helpers.open_gui')



'''
  continuity is a condition for internail
  boundary. 
  It has no contribution to weak form. This should be
  fine as far as mesh is conforming.
'''
from petram.model import Domain, Bdry, Pair
from petram.phys.phys_model  import Phys

class PhysContinuity(Bdry, Phys):
    is_essential = False
    def __init__(self, **kwargs):
        super(PhysContinuity, self).__init__( **kwargs)
        Phys.__init__(self)
        self.sel_readonly = False
        self.sel_index = []



import numpy as np
from collections import defaultdict
from scipy.sparse import lil_matrix, coo_matrix

def fill_table(table1, table2, i, j, value):
    if i > j:
        if table1[i, j] == 0: 
            table1[i, j] = value
        elif table1[i, j] == value:  return
        else:
            if table1[i, j] > value:
                table2[i, j] = table1[i, j]
                table1[i, j] = value                
            else:
                table2[i, j] = value            
    else:
        if table1[j, i] == 0: 
            table1[j, i] = value
        elif table1[j, i] == value:  return            
        else:
            if table1[j, i] > value:
                table2[j, i] = table1[j, i]
                table1[j, i] = value                
            else:
                table2[j, i] = value            
            
def find_edges(mesh):

    iv = mesh.GetBdrElementVertices(0)
    l = len(iv)
    ne = mesh.GetNEdges()        

    if l == 2:
        # 2D mesh
        get_edges = mesh.GetElementEdges
        get_attr  = mesh.GetAttribute
        iattr= mesh.GetAttributeArray()     # min of this array is 1
        nattr = np.max(iattr)
        nb = mesh.GetNE()        
    else:
        # 3D mesh
        get_edges = mesh.GetBdrElementEdges
        get_attr  = mesh.GetBdrAttribute
        iattr= mesh.GetBdrAttributeArray()  # min of this array is 1
        nattr = np.max(iattr)        
        nb = mesh.GetNBE()

    edges = defaultdict(list)
    for i in range(nb):
        ie, io = get_edges(i)
        iattr = get_attr(i)
        edges[iattr].extend(list(ie))
        
    # for each iattr real edge appears only once
    for key in edges.keys():
        seen = defaultdict(int)
        for x in edges[key]: seen[x] +=1
        edges[key] = [k for k in seen if seen[k] == 1]
        
    N = np.hstack([np.zeros(len(edges[k]))+k-1 for k in edges.keys()])
    M = np.hstack([np.array(edges[k]) for k in edges.keys()])
    data = M*0+1
                  
    table1 = coo_matrix((data, (M, N)), shape = (ne, nattr), dtype = int)
    csr = table1.tocsr()

    #embeded surface only touches to one iattr
    idx = np.where(np.diff(csr.indptr) >= 1)[0]
    csr = csr[idx, :]    

    # this is true bdr edges.
    bb_edges = defaultdict(list)
    indptr = csr.indptr; indices = csr.indices
    
    for i in range(csr.shape[0]):
        idxs = tuple(indices[indptr[i]:indptr[i+1]]+1)
        bb_edges[idxs].append(idx[i])
    bb_edges.default_factory = None

    '''
    edges : face index -> edge elements
    bb_edges : set of face index -> edge elements
    '''
    return edges, bb_edges

def plot_edges(mesh, face = 'all'):
    '''
    plot edges between boundary and boundary
       bb : 'all' : plot all edges
             35   : plot edges of battr = 35 as one solid object
    '''
    
    from ifigure.interactive import figure
    
    edges, bb_edges = find_edges(mesh)

    if face == 'all':
        battrs = np.arange(len(edges), dtype=int)
    else:
        battrs = [face-1]
        
    viewer = figure()
    viewer.threed('on')

    for battr in battrs:
        iedges = edges[battr]
        verts = np.stack([np.vstack((mesh.GetVertexArray(mesh.GetEdgeVertices(ie)[0]),
                                     mesh.GetVertexArray(mesh.GetEdgeVertices(ie)[1])))
                                     for ie in iedges])
        viewer.solid(verts)
        return iedges

def plot_bbedges(mesh, bb = 'all'):
    '''
    plot edges between boundary and boundary
       bb : 'all' : plot all edges
             35   : plot edges of battr = 35 splitted based on boundary connection
             (35 36): plot edges between  battr = 35 and 36

    '''
    from ifigure.interactive import figure
    
    edges, bb_edges = find_edges(mesh)
    bb_bdrs = bb_edges.keys()
    
    viewer = figure()
    viewer.threed('on')

    for bb_bdr in bb_bdrs:
        if bb != 'all':
            if isinstance(bb, tuple):
                if any([not x in bb_bdr for x in bb]): continue
            else:
               if not bb in  bb_bdr: continue
        
        #if bb != 'all' and bb != bb_bdr: continue
        
        iedges = bb_edges[bb_bdr]
        verts = np.stack([np.vstack((mesh.GetVertexArray(mesh.GetEdgeVertices(ie)[0]),
                                     mesh.GetVertexArray(mesh.GetEdgeVertices(ie)[1])))
                                     for ie in iedges])
        viewer.solid(verts)
    
    
    



def call_glvis(solpath = None,
               np  = -1,
               thread = True):

    import subprocess as sp
    import sys
    import mfem
    import os
    from threading  import Thread
    import time

    try:
        from Queue import Queue, Empty
    except ImportError:
        from queue import Queue, Empty  # python 3.x

    ON_POSIX = 'posix' in sys.builtin_module_names

    def enqueue_output(out, queue):
       for line in iter(out.readline, b''):
            queue.put(line)
       out.close()
    od = os.getcwd()
    if solpath is None: 
       solpath = model.param.eval('sol').owndir()
    os.chdir(solpath)
    args = ['glvis']
    if np < 0: 
        files = os.listdir(solpath)
        mfiles = [x for x in files if x.startswith('solmesh')]
        if mfiles[0] != 'solmesh': ## if not serial run result
#        args.extend([str(x) for x in globals()['default_glvis_args']])
            args.extend(['-np', str(len(mfiles))])
    else:
        args.extend(['-np', str(np)])
    args.extend(['-m', 'solmesh',  '-g',  'solr', '-gc', '2'])
    print(args)
    p = sp.Popen(args, stdout=sp.PIPE, stderr=sp.STDOUT)
    os.chdir(od)
    if thread:
        q = Queue()
        t = Thread(target=enqueue_output, args=(p.stdout, q))
        t.daemon = True # thread dies with the program
        t.start()
        while(t.is_alive()):
           try:
               line = q.get_nowait() # or q.get(timeout=.1)
           except Empty:
               time.sleep(1.0)
               pass #print('no output yet')
           else: 
               print(line.rstrip('\r\n'))
    else:
        stdoutdata, stderrdata = p.communicate()
        print(stdoutdata)
  
ans(call_glvis(*args, **kwargs))



__path__ = __import__('pkgutil').extend_path(__path__, __name__)
#import os
#path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
#from mfem_model import MFEM_ModelRoot, MFEM_GeneralRoot, MFEM_MeshRoot, MFEM_PhysRoot, MFEM_SolverRoot



'''
   BdrNodalEvaluator:
      a thing to evaluate solution on a boundary
'''
import numpy as np
import weakref
import six

from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
    from mfem.par import GlobGeometryRefiner as GR
else:
    import mfem.ser as mfem
    from mfem.ser import GlobGeometryRefiner as GR

from petram.sol.evaluator_agent import EvaluatorAgent
Geom = mfem.Geometry()


def process_iverts2nodals(mesh, iverts):
    ''' 
    collect data to evalutate nodal values of mesh
    at iverts
    '''
    # we dont want to process the same vert many times.
    # so first we take a unique set...
    iverts_f, iverts_inv = np.unique(iverts.flatten(),
                                     return_inverse=True)
    iverts_inv = iverts_inv.reshape(iverts.shape)

    # then get unique set of elements relating to the verts.
    vert2el = mesh.GetVertexToElementTable()
    ieles = np.hstack([vert2el.GetRowList(i) for i in iverts_f])
    ieles = np.unique(ieles)

    # map from element -> (element's vert index, ivert_f index)
    elvert2facevert = [None]*len(ieles)
    elvertloc = [None]*len(ieles)
    elattr = [None]*len(ieles)

    wverts = np.zeros(len(iverts_f))
    for kk, iel in enumerate(ieles):
        elvert2facevert[kk] = []
        elvertloc[kk] = []
        elattr[kk] = mesh.GetAttribute(iel)
        elverts = mesh.GetElement(iel).GetVerticesArray()
        for k, elvert in enumerate(elverts):
            idx = np.searchsorted(iverts_f, elvert)
            if idx == len(iverts_f):
                continue  # not found
            if iverts_f[idx] != elvert:
                continue  # not found
            elvert2facevert[kk].append((k, idx))
            elvertloc[kk].append(mesh.GetVertexArray(elvert))
            wverts[idx] = wverts[idx]+1

    # idx of element needs to be evaluated

    return {'ieles': np.array(ieles),
            'elvert2facevert': elvert2facevert,
            'locs': np.stack([mesh.GetVertexArray(k) for k in iverts_f]),
            'elvertloc': elvertloc,
            'elattr': np.array(elattr),
            'iverts_inv': iverts_inv,
            'iverts_f': iverts_f,
            'wverts': wverts}


def edge_detect(index):
    #print("edge_detect", index.shape)
    store = []

    def check_pair(store, a, b):
        a1 = min(a, b)
        b1 = max(a, b)
        p = (a1, b1)
        if p in store:
            store.remove(p)
        else:
            store.append(p)
        return store

    for iv in index:
        store = check_pair(store, iv[0], iv[1])
        store = check_pair(store, iv[0], iv[2])
        store = check_pair(store, iv[1], iv[2])
    ret = np.vstack(store)
    return ret


def _get_emesh_idx(obj, expr, solvars, phys, default):
    from petram.helper.variables import Variable, var_g, NativeCoefficientGenBase

    code = compile(expr, '<string>', 'eval')
    names = code.co_names

    g = {}
    # print solvars.keys()
    # print phys._global_ns.keys()
    for key in phys._global_ns.keys():
        g[key] = phys._global_ns[key]
    for key in solvars.keys():
        g[key] = solvars[key]

    idx = []

    for n in names:
        if ((n in g and isinstance(g[n], NativeCoefficientGenBase)) or
                (n in g and isinstance(g[n], Variable))):
            for nn in g[n].dependency:
                idx = g[nn].get_emesh_idx(idx, g=g)

            idx.extend(g[n].get_emesh_idx(idx, g=g))

    if len(idx) == 0:
        # if expression has no emehs dependence return default.
        idx = [default]
    return list(set(idx))


def get_emesh_idx(obj, exprs, solvars, phys):
    '''
    handle multiple expressions and find unique set of emesh_idx needed.
    '''
    if not isinstance(exprs, list):
        exprs = [exprs]

    default = -1
    emesh_idx_list = []
    for item in exprs:
        emesh_idx_list.extend(_get_emesh_idx(
            obj, item, solvars, phys, default))
        default = emesh_idx_list[-1]

    ret = list(set(emesh_idx_list))

    if len(ret) == 1 and ret[0] == -1:
        return [0]
    else:
        ret = [x for x in ret if x != -1]
        return ret


def eval_at_nodals(obj, expr, solvars, phys, edge_evaluator=False):
    '''
    evaluate nodal valus based on preproceessed 
    geometry data

    to be done : obj should be replaced by a dictionary
    '''

    from petram.helper.variables import (Variable,
                                         var_g,
                                         NativeCoefficientGenBase,
                                         CoefficientVariable,
                                         NumbaCoefficientVariable,)

    if len(obj.iverts) == 0:
        return None
    variables = []

    code = compile(expr, '<string>', 'eval')
    names = code.co_names

    g = {}

    for key in phys._global_ns.keys():
        g[key] = phys._global_ns[key]
    for key in solvars.keys():
        g[key] = solvars[key]

    ll_name = []
    ll_value = []
    var_g2 = var_g.copy()

    new_names = []
    name_translation = {}

    all_names = list(names[:])

    def get_names(names):
        for n in names:
            if (n in g and isinstance(g[n], Variable)):
                new_names = g[n].get_names()
                for x in new_names:
                    all_names.append(x)
                get_names(new_names)
    get_names(names)

    for n in all_names:
        if (n in g and isinstance(g[n], NativeCoefficientGenBase)):
            g[n+"_coeff"] = CoefficientVariable(g[n], g)
            new_names.append(n+"_coeff")
            name_translation[n+"_coeff"] = n

        elif (n in g and isinstance(g[n], NumbaCoefficientVariable)):
            ind_vars = [xx.strip() for xx in phys.ind_vars.split(',')]
            g[n].set_coeff(ind_vars, g)
            new_names.append(n)
            name_translation[n] = n

        elif (n in g and isinstance(g[n], Variable)):
            new_names.extend(g[n].dependency)
            new_names.append(n)
            name_translation[n] = n

        elif n in g:
            new_names.append(n)
            name_translation[n] = n

    for n in new_names:
        if (n in g and isinstance(g[n], Variable)):
            if not g[n] in obj.knowns:
                obj.knowns[g[n]] = (
                    g[n].nodal_values(iele=obj.ieles,
                                      ibele=obj.ibeles,
                                      elattr=obj.elattr,
                                      el2v=obj.elvert2facevert,
                                      locs=obj.locs,
                                      elvertloc=obj.elvertloc,
                                      wverts=obj.wverts,
                                      mesh=obj.mesh()[obj.emesh_idx],
                                      iverts_f=obj.iverts_f,
                                      g=g,
                                      knowns=obj.knowns,
                                      edge_evaluator=edge_evaluator))

            ll_name.append(name_translation[n])
            ll_value.append(obj.knowns[g[n]])
        elif (n in g):
            var_g2[n] = g[n]

    if len(ll_value) > 0:
        val = np.array([eval(code, var_g2, dict(zip(ll_name, v)))
                        for v in zip(*ll_value)])
    else:
        # if expr does not involve Varialbe, evaluate code once
        # and generate an array
        val = np.array([eval(code, var_g2)]*len(obj.locs))

    return val


class BdrNodalEvaluator(EvaluatorAgent):
    def __init__(self, battrs, decimate=1):
        super(BdrNodalEvaluator, self).__init__()
        self.battrs = battrs
        self.decimate = decimate

    def preprocess_geometry(self, battrs, emesh_idx=0, decimate=1):

        mesh = self.mesh()[emesh_idx]
        # print 'preprocess_geom',  mesh, battrs
        self.battrs = battrs

        self.decimate = decimate
        self.knowns = WKD()
        self.iverts = []

        if mesh.Dimension() == 3:
            getarray = mesh.GetBdrArray
            getelement = mesh.GetBdrElement
        elif mesh.Dimension() == 2:
            getarray = mesh.GetDomainArray
            getelement = mesh.GetElement
        else:
            assert False, "BdrNodal Evaluator is not supported for this dimension"

        x = [getarray(battr) for battr in battrs]
        if np.sum([len(xx) for xx in x]) == 0:
            return

        ibdrs = np.hstack(x).astype(int).flatten()

        if self.decimate != 1:
            ibdrs = ibdrs[::self.decimate]

        self.ibeles = np.array(ibdrs)

        def get_vertices_array(i):
            arr = getelement(i).GetVerticesArray()
            if len(arr) == 3:
                return arr
            elif len(arr) == 4:
                x = arr[:-1]
                y = np.array([arr[0], arr[2], arr[3]])
                return np.vstack([x, y])

        # we handle quad as two triangles
        iverts = np.vstack([get_vertices_array(i) for i in ibdrs])

        self.iverts = iverts
        if len(self.iverts) == 0:
            return

        data = process_iverts2nodals(mesh, iverts)
        for k in list(data):
            setattr(self, k, data[k])
        self.emesh_idx = emesh_idx

    def eval(self, expr, solvars, phys, **kwargs):
        emesh_idx = get_emesh_idx(self, expr, solvars, phys)
        if len(emesh_idx) > 1:
            assert False, "expression involves multiple mesh (emesh length != 1)"
        # if len(emesh_idx) < 1:
        #    assert False, "expression is not defined on any mesh"
        # (this could happen when expression is pure geometryical like "x+y")
        decimate = kwargs.pop('decimate', 1)

        if len(emesh_idx) == 1:
            if self.emesh_idx != emesh_idx[0]:
                #print("process geom", emesh_idx[0])
                self.preprocess_geometry(self.battrs,
                                         emesh_idx=emesh_idx[0],
                                         decimate=decimate)

        val = eval_at_nodals(self, expr, solvars, phys)

        if val is None:
            return None, None, None

        edge_only = kwargs.pop('edge_only', False)
        export_type = kwargs.pop('export_type', 1)
        if export_type == 2:
            return self.locs, val, None

        refine = kwargs.pop('refine', 1)

        if refine == 1 or edge_only:
            if not edge_only:
                return self.locs, val, self.iverts_inv
            else:
                idx = edge_detect(self.iverts_inv)
                return self.locs, val, idx
        else:
            from petram.sol.nodal_refinement import refine_surface_data

            ptx, data, ridx = refine_surface_data(self.mesh()[self.emesh_idx],
                                                  self.ibeles,
                                                  val, self.iverts_inv,
                                                  refine)
            return ptx, data, ridx



import os
import wx
from ifigure.utils.edit_list import EditListPanel, EDITLIST_CHANGED

setting1 = {"style": wx.CB_READONLY, "choices": (
    "None", "Begin", "End", "Fail", "All")}


def elp_setting(log_keywords):
    ll = [["Num. of Nodes", 1, 400, {}],

          ["Num. of Cores(total)", 16, 400, {}],
          ["Num. of OMP threads", 4, 400, {}],
          ["Wall clock", "00:15:00", 0, {}],
          ["Queue(Repo.)", "Debug", 0, {}],
          ["PetraM Ver.", "19700521", 0, {}],
          ["Remote Dir.", "", 0, {}],
          ["Note", None, 2235, {'nlines': 3}],
          ["Keywords", None, 36, {'col': 3, 'labels': list(log_keywords)}],
          ["Notification", "None", 4, setting1],
#          ["-> Adv. options", None, None, {"no_tlw_resize": True}],
          ["-> Adv. options", None, None, {"tlb_resize_samewidth": True}],
#          ["-> Adv. options", None, None, {"no_tlw_resize": False}],
          ["Run ", None, 2235, {'nlines': 3}],
          ["Env. ", None, 2235, {'nlines': 2}],
          ["<-"],
          [None, False, 3, {"text": "Skip sending mesh file"}],
          ]
    return ll


values = ['1', '1', '1', '00:10:00', 'regular(PROJ_19700521)', '', '',
          '', '', "None", '', '', False, False, ]

keys = ['num_nodes', 'num_cores', 'num_openmp', 'walltime',
        'queue', 'petramver', 'rwdir',
        'log_txt', 'log_keywords', 'notification', 'adv_opts', 'env_opts', 'skip_mesh',
        'retrieve_files']

def_queues = {'type': 'SLURM',
              'queus': [{'name': 'debug',
                         'maxnode': 1}]}


def get_defaults():
    return values[:], keys[:]


class dlg_jobsubmission(wx.Dialog):
    def __init__(self, parent, id=wx.ID_ANY, title='',
                 value=None, queues=None):

        if queues is None:
            queues = def_queues

        wx.Dialog.__init__(self, parent, wx.ID_ANY, title,
                           style=wx.STAY_ON_TOP | wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        vbox = wx.BoxSizer(wx.VERTICAL)
        hbox2 = wx.BoxSizer(wx.HORIZONTAL)

        self.SetSizer(vbox)
        vbox.Add(hbox2, 1, wx.EXPAND | wx.ALL, 1)

        q_names = [x['name'] for x in queues['queues']]
        dp_setting = {"style": wx.CB_DROPDOWN, "choices": q_names}

        v_names = list(queues['versions'])
        dv_setting = {"style": wx.CB_DROPDOWN, "choices": v_names}
        if "keywords" in queues:
            log_keywords = queues["keywords"]
        else:
            log_keywords = ["production", "debug"]
        ll = elp_setting(log_keywords)

        ll[4] = [ll[4][0], q_names[0], 4, dp_setting]
        ll[5] = [ll[5][0], v_names[0], 4, dv_setting]

        self.elp = EditListPanel(self, ll)

        hbox2.Add(self.elp, 1, wx.EXPAND | wx.RIGHT | wx.LEFT, 1)
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        button = wx.Button(self, wx.ID_ANY, "Cancel")
        button2 = wx.Button(self, wx.ID_ANY, "Submit")

        hbox.Add(button, 0, wx.EXPAND)
        hbox.AddStretchSpacer()
        hbox.Add(button2, 0, wx.EXPAND)
        vbox.Add(hbox, 0, wx.EXPAND | wx.ALL, 5)

        button.Bind(wx.EVT_BUTTON, self.onCancel)
        button2.Bind(wx.EVT_BUTTON, self.onSubmit)

        #print("value", value)
        petram_version = value[4].split("_")[1][:-1]
        queue = value[4].split("_")[0]+")"
        if value is not None:
            v, names = get_defaults()
            for k, n in enumerate(names):
                if n in value:
                    value[n] = v[k]

            value[4] = queue
            value[5] = petram_version
            if not value[4] in q_names:
                value[4] = q_names[0]
            if not value[5] in v_names:
                value[5] = v_names[0]

            value8 = [False]*len(log_keywords)
            for name, v in value[8]:
                if name in log_keywords:
                    value8[log_keywords.index(name)] = v
            value[8] = value8

            tmp = [y for x, y in queues['versions']
                   [value[5]] if x == 'srun_option']
            value[10] = '\n'.join(tmp)
            tmp = [y for x, y in queues['versions']
                   [value[5]] if x == 'env_option']
            value[11] = '\n'.join(tmp)

            self.elp.SetValue(value)

        #self.SetSizeHints(minH=-1, minW=size.GetWidth())
        self.SetSizeHints(minH=-1, minW=300)

        self.Layout()
        self.Fit()
        size = self.GetSize()
        width = max(len(value[5])*12, 550)
        width = min(width, 1200)
        self.SetSize((width, size.GetHeight()))
        self.CenterOnParent()

        self.Show()
        # wx.CallAfter(self.Fit)

        self.value = self.elp.GetValue()
        self._queues = queues
        self.Bind(EDITLIST_CHANGED, self.onEL_Changed)

        user_notice = "\n".join(queues["notice"])
        from ifigure.widgets.dialog import message

        ret = message(parent=self,
                      message=user_notice,
                      title='Notice',
                      style=2,
                      icon=wx.ICON_EXCLAMATION,
                      center_on_screen=False,
                      center_on_parent=True,
                      labels=["Accept", "Cancel"])

        if ret == "cancel":
            wx.CallAfter(self.onCancel)

    def onCancel(self, evt=None):
        self.value = self.elp.GetValue()
        self.EndModal(wx.ID_CANCEL)

    def onSubmit(self, evt):
        self.value = self.elp.GetValue()

        if (self.value[7].strip() == '' or
                not any([x[1] for x in self.value[8]])):

            from ifigure.widgets.dialog import message
            message(self, title="Error",
                    message="Enter job description and select at least one keyword")
            return
        self.EndModal(wx.ID_OK)
        evt.Skip()

    def onEL_Changed(self, evt):
        # update adv. options if Petra-M
        value = self.elp.GetValue()

        if self.value[5] != value[5]:
            # need to update values for buttons
            value[8] = [y for x, y in value[8]]

            # update adv. options.
            #tmp = [y for x, y in self._queues['versions'][value[5]]]
            #value[10] = '\n'.join(tmp)
            tmp = [y for x, y in self._queues['versions']
                   [value[5]] if x == 'srun_option']
            value[10] = '\n'.join(tmp)
            tmp = [y for x, y in self._queues['versions']
                   [value[5]] if x == 'env_option']
            value[11] = '\n'.join(tmp)

            self.elp.SetValue(value)
            self.value = value

        evt.Skip()


def get_job_submisson_setting(parent, servername='', value=None,
                              queues=None):

    if value[6] == '':
        from petram.remote.client_script import wdir_from_datetime
        value[6] = wdir_from_datetime()
    else:
        value[6] = os.path.basename(value[6])

    dlg = dlg_jobsubmission(parent, title='Submit to '+servername, value=value,
                            queues=queues)

    value = {}
    base_remote_path = queues['scratch']
    try:
        if dlg.ShowModal() == wx.ID_OK:
            value["num_nodes"] = dlg.value[0]
            value["num_cores"] = dlg.value[1]
            value["num_openmp"] = dlg.value[2]
            value["walltime"] = str(dlg.value[3])

            queue_value = str(dlg.value[4])[:-1]+'_'+str(dlg.value[5])+")"
            value["queue"] = queue_value
            value["retrieve_files"] = False
            value["rwdir"] = os.path.join(base_remote_path, dlg.value[6])
            value["log_txt"] = dlg.value[7]
            value["log_keywords"] = dlg.value[8]
            value["notification"] = dlg.value[9]
            value["adv_opts"] = dlg.value[10]
            value["env_opts"] = dlg.value[11]
            value["skip_mesh"] = dlg.value[12]
        else:
            pass
    finally:
        dlg.Destroy()
    return value



def clear_sol(w = proj.app):
      for name, child in model.solutions.get_children():
           child.destroy()
      model.param.delvar('sol')   
      from ifigure.events import SendChangedEvent
      SendChangedEvent(model, w=w)
ans(clear_sol())



def run_parallel(path='', nproc=1, debug=0, thread=True):
    '''
    debug keyword will overwrite debug level setting 
    in model file
    '''
    import subprocess as sp
    import sys
    import mfem
    import os
    from threading import Thread
    import time

    try:
        from Queue import Queue, Empty
    except ImportError:
        from queue import Queue, Empty  # python 3.x

    ON_POSIX = 'posix' in sys.builtin_module_names

    def enqueue_output(out, queue):
        for line in iter(out.readline, b''):
            queue.put(line)
        out.close()

    del_path = False
    if path == '':
        if model.param.eval('sol') is None:
            folder = model.scripts.helpers.make_new_sol()
        else:
            folder = model.param.eval('sol')
            folder.clean_owndir()
        path = os.path.join(folder.owndir(), 'model.pmfm')
        model.scripts.helpers.save_model(path)
        m = model.param.getvar('mfem_model')
        try:
            m.generate_script(dir=folder.owndir())
        except:
            import traceback
            traceback.print_exc()
            return
        del_path = True

    import petram
    #from petram.helper.driver_path import parallel as driver
    opath = os.getcwd()

    #args = ['mpirun', '-n', str(nproc), driver, str(path),  str(debug)]
    args = ['mpirun', '-n', str(nproc), sys.executable, '-u',
            'model.py', '-p', '-d',  str(debug)]
    os.chdir(folder.owndir())
    p = sp.Popen(args, stdout=sp.PIPE, stderr=sp.STDOUT)

    if thread:
        q = Queue()
        t = Thread(target=enqueue_output, args=(p.stdout, q))
        t.daemon = True  # thread dies with the program
        t.start()

        while(True):
            try:
                line = q.get_nowait()  # or q.get(timeout=.1)
            except Empty:
                if not t.is_alive():
                    break
                time.sleep(1.0)
                pass  # print('no output yet')
            else:
                if isinstance(line, bytes):
                    line = line.decode('utf-8')
                    line = '\n'.join(
                        [x for x in line.split('\n') if len(x) > 0])
                print(line)
    else:
        stdoutdata, stderrdata = p.communicate()
        print(stdoutdata)

    p.kill()
    globals()['default_sol_path'] = os.path.dirname(path)
    globals()['default_glvis_args'] = ['-np', nproc]

    from petram.sol.solsets import read_sol, find_solfiles
    path = model.param.eval('sol').owndir()
    try:
        solfiles = find_solfiles(path=path)
        model.variables.setvar('solfiles', solfiles)
    except:
        model.variables.delvar('solfiles')
    #if del_path: os.remove(path)

    os.chdir(opath)


ans(run_parallel(*args, **kwargs))



from petram.phys.phys_model import (PhysConstant,
                                    PhysVectorConstant,
                                    PhysMatrixConstant)
from petram.helper.variables import NativeCoefficientGenBase

import weakref
import numpy as np

from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    comm = MPI.COMM_WORLD
    from petram.helper.mpi_recipes import gather_vector, allgather_vector
    from mfem.common.mpi_debug import nicePrint
else:
    import mfem.ser as mfem

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints("Operators")
rprint = debug.regular_print('Operators')


class Operator(object):
    def __repr__(self):
        return self.__class__.__name__ + "("+",".join(self.sel)+")"


class Operator(object):
    def __init__(self, **kwargs):
        self._sel = kwargs.pop("sel", "all")
        self._ssel = kwargs.pop("src", "all")
        self._sel_mode = kwargs.pop("sel_mode", "domain")
        self._fes1 = None
        self._fes2 = None
        self._engine = None
        self._transpose = False
        self._trial_ess_tdof = None
        self._test_ess_tdof = None
        self._c_coeff = None   # coefficient

    def __call__(self, *args, **kwargs):
        return self.assemble(*args, **kwargs)

    def assemble(self, fes):
        raise NotImplementedError("Subclass needs to implement this")

    def process_kwargs(self, engine, kwargs):
        '''
        kwargs in expression can overwrite this.
        '''
        fes1 = kwargs.pop('test', None)
        fes2 = kwargs.pop('trial', None)
        self._transpose = kwargs.pop('transpose', self._transpose)
        self._sel_mode = kwargs.pop('sel_mode', self._sel_mode)
        self._sel = kwargs.pop('sel', self._sel)
        self._ssel = kwargs.pop('src', self._ssel)
        if fes1 is not None:
            if isinstance(fes1, str):
                fes1 = engine.fespaces[fes]
            self._fes1 = weakref.ref(fes1)
        if fes2 is not None:
            if isinstance(fes2, str):
                fes2 = engine.fespaces[fes]
            self._fes2 = weakref.ref(fes2)

    @property
    def fes1(self):
        return self._fes1()

    @property
    def fes2(self):
        return self._fes2()

    @property
    def sel_mode(self):
        return self._sel_mode

    def convert_mat_to_operator(self, mat):
        '''
        a utility routine to convert locally assembled mat
        to linear operator
        '''
        is_complex = np.iscomplexobj(mat)
        m_coo = mat.tocoo()
        row = m_coo.row
        col = m_coo.col
        col = np.unique(col)

        from scipy.sparse import coo_matrix, csr_matrix
        if use_parallel:
            if is_complex:
                m1 = csr_matrix(mat.real, dtype=float)
                m2 = csr_matrix(mat.imag, dtype=float)
            else:
                m1 = csr_matrix(mat.real, dtype=float)
                m2 = None
            from mfem.common.chypre import CHypreMat
            start_col = self.fes1.GetMyTDofOffset()
            end_col = self.fes1.GetMyTDofOffset() + self.fes1.GetTrueVSize()
            col_starts = [start_col, end_col, mat.shape[1]]

            M = CHypreMat(m1, m2, col_starts=col_starts)
        else:
            from petram.helper.block_matrix import convert_to_ScipyCoo

            M = convert_to_ScipyCoo(coo_matrix(mat, dtype=mat.dtype))


class LoopIntegral(Operator):
    def assemble(self, *args, **kwargs):
        '''
        looplintegral(18)   # integrate around the bdr 18

        # this mode is only avaiable on serial version.
        looplintegral([1,2,3], [4, 5]) # loop defined by two groups of boundaries 
        '''
        coeff = kwargs.pop("coeff", "1")
        coeff_type = kwargs.pop("coeff_type", "S")

        engine = self._engine()
        self.process_kwargs(engine, kwargs)

        face = args

        if not self.fes1.FEColl().Name().startswith('ND'):
            assert False, "line integration is only implemented for ND"

        fes1 = self.fes1
        info1 = engine.get_fes_info(self.fes1)
        emesh_idx = info1['emesh_idx']
        mesh = engine.emeshes[emesh_idx]

        if use_parallel:
            from petram.mesh.find_loop import find_loop_par
            idx, signs = find_loop_par(mesh, *face)
            fesize1 = fes1.GetTrueVSize()
            P = fes1.Dof_TrueDof_Matrix()
            from mfem.common.parcsr_extra import ToScipyCoo
            P = ToScipyCoo(P).tocsr()
            VDoFtoGTDoF = P.indices
            rstart = fes1.GetMyTDofOffset()
        else:
            from petram.mesh.find_loop import find_loop_ser
            idx, signs = find_loop_ser(mesh, *face)
            fesize1 = fes1.GetNDofs()

        map = np.zeros((fesize1, 1), dtype=float)

        w = []
        for sign, ie in zip(signs, idx):
            dofs = fes1.GetEdgeDofs(ie)
            # don't put this Tr outside the loop....
            Tr = mesh.GetEdgeTransformation(ie)
            weight = Tr.Weight()
            w.append(Tr.Weight())
            for dof in dofs:
                if use_parallel:
                    dof = VDoFtoGTDoF[dof] - rstart
                map[dof] = sign

        #if len(w) > 0: print("weight", w)

        from mfem.common.chypre import PyVec2PyMat, CHypreVec
        if use_parallel:
            v1 = CHypreVec(map.flatten(), None)
        else:
            v1 = map.reshape((-1, 1))

        v1 = PyVec2PyMat(v1)
        if not self._transpose:
            v1 = v1.transpose()
        return v1


class Integral(Operator):
    def assemble(self, *args, **kwargs):
        '''
        integral()
        integral('boundary', [1,3])  # selection type and selection
        integral('boundary', [1,3], integrator = 'auto' or other MFEM linearform integrator
        integral('boundary', [1,3], weight = '1')

        '''

        coeff = kwargs.pop("coeff", "1.0")
        coeff_type = kwargs.pop("coeff_type", "S")

        engine = self._engine()
        self.process_kwargs(engine, kwargs)

        if len(args) > 0:
            self._sel_mode = args[0]
        if len(args) > 1:
            self._sel = args[1]

        if (self.fes1.FEColl().Name().startswith('ND') or
                self.fes1.FEColl().Name().startswith('RT')):
            cdim = self.fes1.GetMesh().SpaceDimension()
        else:
            cdim = self.fes1.GetVDim()

        info1 = engine.get_fes_info(self.fes1)
        emesh_idx = info1['emesh_idx']

        isDomain = (self._sel_mode == 'domain')

        from petram.helper.phys_module_util import default_lf_integrator
        integrator = kwargs.pop("integrator", "Auto")
        if integrator == 'Auto':
            integrator = default_lf_integrator(info1, isDomain)

        global_ns = globals()
        local_ns = {}
        real = True

        try:
            is_complex = np.iscomplex(
                complex(eval(coeff), global_ns, local_ns))
        except:
            is_complex = kwargs.pop('is_complex', False)

        if isDomain:
            adder = 'AddDomainIntegrator'
        else:
            adder = 'AddBoundaryIntegrator'

        from petram.helper.phys_module_util import restricted_integrator
        from mfem.common.chypre import LF2PyVec, PyVec2PyMat, MfemVec2PyVec

        itg = restricted_integrator(engine, integrator, self._sel,
                                    coeff, coeff_type, cdim,
                                    emesh_idx,
                                    isDomain,
                                    self._ind_vars, local_ns, global_ns, True)

        lf1 = engine.new_lf(self._fes1())
        getattr(lf1, adder)(itg)
        lf1.Assemble()

        if is_complex:
            itg = restricted_integrator(engine, integrator, self._sel,
                                        coeff, coeff_type, cdim,
                                        emesh_idx, isDomain,
                                        self._ind_vars, local_ns, global_ns, False)

            lf2 = engine.new_lf(self._fes1())
            getattr(lf2, adder)(itg)
            lf2.Assemble()
            v1 = MfemVec2PyVec(engine.b2B(lf1), engine.b2B(lf2))
        else:
            v1 = MfemVec2PyVec(engine.b2B(lf1), None)
            lf2 = None

        v1 = PyVec2PyMat(v1)
        if not self._transpose:
            v1 = v1.transpose()
        return v1


def make_diagonal_mat(engine, fes1, fes2, value):
    if fes1 == fes2:
        bf = engine.new_bf(fes1)
        bf.Assemble()
        bf.Finalize()
        mat = engine.a2A(bf)
    else:
        bf = engine.new_mixed_bf(fes1, fes2)
        #one = mfem.ConstantCoefficient(0.0001)
        #itg = mfem.MixedScalarMassIntegrator()
        # bf.AddDomainIntegrator(itg)
        bf.Assemble()
        mat = engine.a2Am(bf)

    if use_parallel:
        mat.CopyRowStarts()
        mat.CopyColStarts()

    from mfem.common.chypre import MfemMat2PyMat
    m1 = MfemMat2PyMat(mat, None)

    if not use_parallel:
        from petram.helper.block_matrix import convert_to_ScipyCoo
        m1 = convert_to_ScipyCoo(m1)
    shape = m1.shape
    assert shape[0] == shape[1], "Identity Operator must be square"

    idx = range(shape[0])
    m1.setDiag(idx, value=value)

    return m1


class Identity(Operator):
    def assemble(self, *args, **kwargs):
        engine = self._engine()
        self.process_kwargs(engine, kwargs)

        fes1 = self._fes1()
        fes2 = fes1 if self._fes2 is None else self._fes2()

        c_coeff = self._c_coeff

        from petram.phys.phys_model import PhysConstant
        coeff = 0.
        if c_coeff[0] is not None:
            assert isinstance(
                c_coeff[0], PhysConstant), "projection supports only constant scalr coefficient"
            coeff += c_coeff[0].value
        if c_coeff[1] is not None:
            assert isinstance(
                c_coeff[1], PhysConstant), "projection supports only constant scalr coefficient"
            coeff += 1j*c_coeff[1].value

        mat = make_diagonal_mat(engine, fes1, fes2, 1.0)

        return mat*coeff

class Zero(Operator):
    def assemble(self, *args, **kwargs):
        engine = self._engine()
        self.process_kwargs(engine, kwargs)

        fes1 = self._fes1()
        fes2 = fes1 if self._fes2 is None else self._fes2()
        return make_diagonal_mat(engine, fes1, fes2, 0.0)

class Delta(Operator):
    '''
    Delta function

      delta(x)
      delta(x, y)
      delta(x, y, z)

      delta(x, y, z, direction=[1,0,0]))
      delta(array) to set multiple points at once
    '''

    def assemble(self, *args, **kwargs):
        engine = self._engine()
        direction = kwargs.pop("direction", 0)

        self.process_kwargs(engine, kwargs)

        x = args[0]
        y = args[1] if len(args) > 1 else 0.0
        z = args[2] if len(args) > 2 else 0.0

        sdim = self.fes1.GetMesh().SpaceDimension()
        if direction == 0:
            if sdim == 3:
                d = mfem.DeltaCoefficient(x, y, z, 1.0)
            elif sdim == 2:
                d = mfem.DeltaCoefficient(x, y, 1.0)
            elif sdim == 1:
                d = mfem.DeltaCoefficient(x, 1.0)
            else:
                assert False, "unsupported dimension"
            intg = mfem.DomainLFIntegrator(d)
        else:
            dir = mfem.Vector(direction)
            if sdim == 3:
                d = mfem.VectorDeltaCoefficient(dir, x, y, z, 1.0)
            elif sdim == 2:
                d = mfem.VectorDeltaCoefficient(dir, x, y, 1.0)
            elif sdim == 1:
                d = mfem.VectorDeltaCoefficient(dir, x, 1.0)
            else:
                assert False, "unsupported dimension"

            if self.fes1.FEColl().Name().startswith('ND'):
                intg = mfem.VectorFEDomainLFIntegrator(d)
            elif self.fes1.FEColl().Name().startswith('RT'):
                intg = mfem.VectorFEDomainLFIntegrator(d)
            else:
                intg = mfem.VectorDomainLFIntegrator(d)

        lf1 = engine.new_lf(self.fes1)
        lf1.AddDomainIntegrator(intg)
        lf1.Assemble()

        from mfem.common.chypre import LF2PyVec, PyVec2PyMat, MfemVec2PyVec
        v1 = MfemVec2PyVec(engine.b2B(lf1), None)

        v1 = PyVec2PyMat(v1)
        if not self._transpose:
            v1 = v1.transpose()
        return v1


class DeltaM(Operator):
    '''
    Delta function

      delta(array, direction = None, weight = None, sum = False) to set multiple points at once

      array = [x1, y1, z1, x2, y2, z2,....]
      direction = [dx1, dy1, dz1, dx2, dy2, dz2,....]
      weight = [w1, w2, w3,...]
      sum = if True, all points are corrapsed to one array.

    '''

    def assemble(self, *args, **kwargs):
        engine = self._engine()
        direction = kwargs.pop("direction", None)
        weight = kwargs.pop("weight", 1.0)
        weight = np.atleast_1d(weight)
        do_sum = kwargs.pop("sum", False)

        info1 = engine.get_fes_info(self.fes1)
        sdim = info1['sdim']
        vdim = info1['vdim']

        if (info1['element'].startswith('ND') or
                info1['element'].startswith('RT')):
            vdim = sdim

        if vdim > 1:
            if direction is None:
                direction = [0]*vdim
                direction[0] = 1
            direction = np.atleast_1d(direction).astype(float, copy=False)
            direction = direction.reshape(-1, sdim)

        self.process_kwargs(engine, kwargs)

        pts = np.array(args[0], copy=False).reshape(-1, sdim)

        from mfem.common.chypre import LF2PyVec, PyVec2PyMat, MfemVec2PyVec, HStackPyVec
        vecs = []

        for k, pt in enumerate(pts):
            w = float(weight[0] if len(weight) == 1 else weight[k])
            if vdim == 1:
                if sdim == 3:
                    x, y, z = pt
                    d = mfem.DeltaCoefficient(x, y, z, w)
                elif sdim == 2:
                    x, y = pt
                    d = mfem.DeltaCoefficient(x, y, w)
                elif sdim == 1:
                    x = pt
                    d = mfem.DeltaCoefficient(x, w)
                else:
                    assert False, "unsupported dimension"
                intg = mfem.DomainLFIntegrator(d)
            else:
                dir = direction[0] if len(direction) == 1 else direction[k]
                dd = mfem.Vector(dir)
                if sdim == 3:
                    x, y, z = pt
                    d = mfem.VectorDeltaCoefficient(dd, x, y, z, w)
                elif sdim == 2:
                    x, y = pt
                    d = mfem.VectorDeltaCoefficient(dd, x, y, w)
                elif sdim == 1:
                    x = pt
                    d = mfem.VectorDeltaCoefficient(dd, x, w)
                else:
                    assert False, "unsupported dimension"

                if self.fes1.FEColl().Name().startswith('ND'):
                    intg = mfem.VectorFEDomainLFIntegrator(d)
                elif self.fes1.FEColl().Name().startswith('RT'):
                    intg = mfem.VectorFEDomainLFIntegrator(d)
                else:
                    intg = mfem.VectorDomainLFIntegrator(d)

            if do_sum:
                if k == 0:
                    lf1 = engine.new_lf(self.fes1)
                lf1.AddDomainIntegrator(intg)
            else:
                lf1 = engine.new_lf(self.fes1)
                lf1.AddDomainIntegrator(intg)
                lf1.Assemble()
                vecs.append(LF2PyVec(lf1))

        if do_sum:
            lf1.Assemble()
            v1 = MfemVec2PyVec(engine.b2B(lf1), None)
            v1 = PyVec2PyMat(v1)
        else:
            v1 = HStackPyVec(vecs)

        if not self._transpose:
            v1 = v1.transpose()
        return v1


class Projection(Operator):
    '''
    DoF mapping (Dof of fes1 is mapped to fes2)

    fes1 is trial space
    fes2 is test  space

    example
    # selection mode is interpreted in the test fes.
    # map all booundary 
    projection("boundary", "all")     
    # map domain 1
    projection("domain", [1])         
    # to give FEspace explicitly
    projection("boundary", [3], trial="u", test="v")
    # to project on a different location
    projection("boundary", [1], src=[2], trans="(x, y, 0)", srctrans="(x, y, 0)")
    '''

    def assemble(self, *args, **kwargs):
        engine = self._engine()
        self.process_kwargs(engine, kwargs)
        if len(args) > 0:
            self._sel_mode = args[0]
        if len(args) > 1:
            self._sel = args[1]

        # transformation of fes1 (or both)
        trans1 = kwargs.pop("trans",   None)
        trans2 = kwargs.pop("srctrans", trans1)  # transformation of fes2
        srcsel = kwargs.pop("src", self._ssel)    # source idx
        tol = kwargs.pop("tol", 1e-5)         # projectio tolerance

        c_coeff = self._c_coeff

        from petram.phys.phys_model import PhysConstant
        coeff = 0.
        if c_coeff[0] is not None:
            assert isinstance(
                c_coeff[0], PhysConstant), "projection supports only constant scalr coefficient"
            coeff += c_coeff[0].value
        if c_coeff[1] is not None:
            assert isinstance(
                c_coeff[1], PhysConstant), "projection supports only constant scalr coefficient"
            coeff += 1j*c_coeff[1].value

        dim1 = self.fes1.GetMesh().Dimension()
        dim2 = self.fes2.GetMesh().Dimension()

        projmode = ""
        if dim2 == 3:
            if self.sel_mode == "domain":
                projmode = "volume"
            elif self.sel_mode == "boundary":
                projmode = "surface"
        elif dim2 == 2:
            if self.sel_mode == "domain":
                projmode = "surface"
            elif self.sel_mode == "boundary":
                projmode = "edge"
        elif dim2 == 1:
            if self.sel_mode == "domain":
                projmode = "edge"
            elif self.sel_mode == "boundary":
                projmode = "vertex"
        assert projmode != "", "unknow projection mode"
        if self._sel == 'all' and self._ssel == 'all':
            if self.sel_mode == "domain":
                if dim1 == dim2:
                    idx1 = np.unique(self.fes1.GetMesh().GetAttributeArray())
                elif dim1 == dim2+1:
                    idx1 = np.unique(
                        self.fes1.GetMesh().GetBdrAttributeArray())
                else:
                    assert False, "unsupported mode"
                idx2 = np.unique(self.fes2.GetMesh().GetAttributeArray())
            else:  # boundary mode
                if dim1 == dim2:
                    idx1 = np.unique(
                        self.fes1.GetMesh().GetBdrAttributeArray())
                elif dim1 == dim2-1:
                    idx1 = np.unique(self.fes1.GetMesh().GetAttributeArray())
                idx2 = np.unique(self.fes2.GetMesh().GetBdrAttributeArray())
            if use_parallel:
                idx1 = list(idx1)
                idx2 = list(idx2)
                idx1 = list(set(sum(comm.allgather(idx1), [])))
                idx2 = list(set(sum(comm.allgather(idx2), [])))
            idx = np.intersect1d(idx1, idx2)
            idx1 = list(idx)
            idx2 = list(idx)

        elif self._ssel == 'all':
            if self.sel_mode == "domain":
                idx2 = np.unique(self.fes2.GetMesh().GetAttributeArray())
            else:
                idx2 = np.unique(self.fes2.GetMesh().GetBdrAttributeArray())
            idx1 = self._sel
            idx2 = list(idx2)
            if use_parallel:
                idx2 = list(set(sum(comm.allgather(idx2), [])))

        elif self._sel == 'all':
            if self.sel_mode == "domain":
                idx1 = np.unique(self.fes1.GetMesh().GetAttributeArray())
            else:
                idx1 = np.unique(self.fes1.GetMesh().GetBdrAttributeArray())

            idx1 = list(idx1)
            if use_parallel:
                idx1 = list(set(sum(comm.allgather(idx1), [])))
            idx2 = srcsel

        else:
            idx1 = self._sel
            idx2 = srcsel

        if use_parallel:
            # we may not need this?
            idx1 = list(set(sum(comm.allgather(idx1), [])))
            idx2 = list(set(sum(comm.allgather(idx2), [])))

        dprint1("projection index ", idx1, idx2)

        from petram.helper.dof_map import notrans

        sdim1 = self.fes1.GetMesh().SpaceDimension()
        sdim2 = self.fes2.GetMesh().SpaceDimension()

        lns = {}
        if trans1 is not None:
            if sdim1 == 3:
                trans1 = ['def trans1(xyz):',
                          '    import numpy as np',
                          '    x = xyz[0]; y=xyz[1]; z=xyz[2]',
                          '    return np.array(['+trans1+'])']
            elif sdim1 == 2:
                trans1 = ['def trans1(xyz):',
                          '    import numpy as np',
                          '    x = xyz[0]; y=xyz[1]',
                          '    return np.array(['+trans1+'])']
            else:  # sdim1 == 3
                trans1 = ['def trans1(xyz):',
                          '    import numpy as np',
                          '    x = xyz[0]',
                          '    return np.array(['+trans1+'])']
            exec('\n'.join(trans1), self._global_ns, lns)
            trans1 = lns['trans1']
        else:
            trans1 = notrans

        if trans2 is not None:
            if sdim2 == 3:
                trans2 = ['def trans2(xyz):',
                          '    import numpy as np',
                          '    x = xyz[0]; y=xyz[1]; z=xyz[2]',
                          '    return np.array(['+trans2+'])']
            elif sdim2 == 2:
                trans2 = ['def trans2(xyz):',
                          '    import numpy as np',
                          '    x = xyz[0]; y=xyz[1]',
                          '    return np.array(['+trans2+'])']
            else:  # sdim1 == 3
                trans2 = ['def trans2(xyz):',
                          '    import numpy as np',
                          '    x = xyz[0]',
                          '    return np.array(['+trans2+'])']

            trans2 = lns['trans2']
        else:
            trans2 = notrans

        from petram.helper.dof_map import projection_matrix as pm
        # matrix to transfer unknown from trail to test
        M, row, col = pm(idx2, idx1, self.fes2, [], fes2=self.fes1,
                         trans1=trans2, trans2=trans1,
                         mode=projmode, tol=tol, filldiag=False,
                         old_mapping=False)
        return M*coeff

# for now we assemble matrix whcih mapps essentials too...
#                        tdof1=self._test_ess_tdof,
#                        tdof2=self._trial_ess_tdof)


class Gradient(Operator):
    '''
    Operator to compute Gradient
    input (domain) should be H1
    output (range) should be ND

    '''

    def assemble(self, *args, **kwargs):
        engine = self._engine()
        self.process_kwargs(engine, kwargs)

        if len(args) > 0:
            self._sel_mode = args[0]
        if len(args) > 1:
            self._sel = args[1]

        if not self.fes1.FEColl().Name().startswith('ND'):
            assert False, "range should be ND"
        if not self.fes2.FEColl().Name().startswith('H1'):
            assert False, "domain should be H1"

        dim1 = self.fes1.GetMesh().Dimension()
        dim2 = self.fes2.GetMesh().Dimension()

        if use_parallel:
            import mfem.par as mfem
            DiscreteLinearOperator = mfem.ParDiscreteLinearOperator
        else:
            import mfem.ser as mfem
            DiscreteLinearOperator = mfem.DiscreteLinearOperator

        grad = DiscreteLinearOperator(self.fes2, self.fes1)
        itp = mfem.GradientInterpolator()
        grad.AddDomainInterpolator(itp)
        grad.Assemble()
        grad.Finalize()

        from mfem.common.chypre import BF2PyMat

        M = BF2PyMat(grad)

        return M


class Curl(Operator):
    '''
    Operator to compute Gradient
    input (domain) should be H1
    output (range) should be ND

    '''

    def assemble(self, *args, **kwargs):
        engine = self._engine()
        self.process_kwargs(engine, kwargs)

        if len(args) > 0:
            self._sel_mode = args[0]
        if len(args) > 1:
            self._sel = args[1]

        # if not self.fes1.FEColl().Name().startswith('ND'):
        #   assert False, "range should be ND"
        if not self.fes2.FEColl().Name().startswith('ND'):
            assert False, "domain should be ND"

        dim1 = self.fes1.GetMesh().Dimension()
        dim2 = self.fes2.GetMesh().Dimension()

        if use_parallel:
            import mfem.par as mfem
            DiscreteLinearOperator = mfem.ParDiscreteLinearOperator
        else:
            import mfem.ser as mfem
            DiscreteLinearOperator = mfem.DiscreteLinearOperator

        curl = DiscreteLinearOperator(self.fes2, self.fes1)
        itp = mfem.CurlInterpolator()
        curl.AddDomainInterpolator(itp)
        curl.Assemble()
        curl.Finalize()

        from mfem.common.chypre import BF2PyMat

        M = BF2PyMat(curl)

        return M


class Divergence(Operator):
    '''
    Operator to compute Gradient
    input (domain) should be RT
    output (range) should be L2

    '''

    def assemble(self, *args, **kwargs):
        engine = self._engine()
        self.process_kwargs(engine, kwargs)

        if len(args) > 0:
            self._sel_mode = args[0]
        if len(args) > 1:
            self._sel = args[1]

        if not self.fes1.FEColl().Name().startswith('L2'):
            assert False, "range should be L2"
        if not self.fes2.FEColl().Name().startswith('RT'):
            assert False, "domain should be RT"

        dim1 = self.fes1.GetMesh().Dimension()
        dim2 = self.fes2.GetMesh().Dimension()

        if use_parallel:
            import mfem.par as mfem
            DiscreteLinearOperator = mfem.ParDiscreteLinearOperator
        else:
            import mfem.ser as mfem
            DiscreteLinearOperator = mfem.DiscreteLinearOperator

        div = DiscreteLinearOperator(self.fes2, self.fes1)
        itp = mfem.DivergenceInterpolator()
        div.AddDomainInterpolator(itp)
        div.Assemble()
        div.Finalize()

        from mfem.common.chypre import BF2PyMat

        M = BF2PyMat(div)

        return M


class Hcurln(Operator):
    '''
    Operator to compute normal compnent of hcurl on surface

    \int F_test(x) coeff(x) F_trial(x') dx'

    input (domain) should be Hcurl
    output (range) should be H1/L2

    Usage: 
       = hcurln(orderinc=1, bdr='all')
    '''

    def assemble(self, *args, **kwargs):
        from petram.helper.hcurl_normal import hcurln

        engine = self._engine()
        verbose = kwargs.pop("verbose", False)
        bdr = kwargs.pop("bdr", 'all')
        orderinc = kwargs.pop("orderinc", 1)
        if bdr != 'all':
            try:
                _void = bdr[0]
            except:
                bdr = [bdr]
        self.process_kwargs(engine, kwargs)

        info1 = engine.get_fes_info(self.fes1)
        emesh1_idx = info1['emesh_idx']
        info2 = engine.get_fes_info(self.fes2)
        emesh2_idx = info2['emesh_idx']

        #assert emesh1_idx == emesh2_idx, "HcurlN is performed only on the same mesh"

        dim1 = self.fes1.GetMesh().Dimension()
        if not self.fes1.FEColl().Name().startswith('ND'):
            assert False, "trial(domain) should be ND"
        if (not self.fes2.FEColl().Name().startswith('L2') and
                not self.fes2.FEColl().Name().startswith('H1')):
            assert False, "test(range) should be H1/L2"

        if dim1 == 3:
            func = hcurln
        elif dim1 == 2:
            func = hcurln
        elif dim1 == 3:
            assert False, "unsupported dimension"
        else:
            assert False, "unsupported dimension"

        M = func(self.fes1,
                 self.fes2,
                 self._c_coeff,
                 is_complex=self._is_complex,
                 bdr=bdr,
                 orderinc=orderinc,
                 verbose=verbose)

        return M


class Convolve(Operator):
    '''
    Operator to compute convolution integral in 1D

    \int F_test(x) coeff(x-x', (x+x')/2.) F_trial(x') dx'

    input (domain) should be H1/L2
    output (range) should be H1

    Usage: 
       = conv1d(coeff, (optional) support, complex=False, 
                orderinc=1, zero_support=False, 
                test_domain='all',
                trial_domain='all')
       coeff is a callable defining the convolution kernel. 
       this function takes two (x-x', x+x'/2) arguments.

       support is a callable, which takes one argment x
       returning the support of kernel at the given location.

       coeff can return None to indicate there is no contribution.
       this can be used as an alternative to using support. In this
       one can check support using (x+x')/2

       The code skips the numerical integration for those points
       sitting outside the support or x-x' > support((x+x')/2.0).

       options:
          orderinc=1 : increase intengration order
          zero_support=False : use zero support function.
    '''

    def assemble(self, *args, **kwargs):
        from petram.helper.convolve import convolve1d, convolve2d, delta, zero

        engine = self._engine()
        is_complex = kwargs.pop("complex", False)
        orderinc = kwargs.pop("orderinc", 1)
        verbose = kwargs.pop("verbose", False)
        zero_support = kwargs.pop("zero_support", False)
        test_domain = kwargs.pop("test_domain", 'all')
        trial_domain = kwargs.pop("trial_domain", 'all')
        self.process_kwargs(engine, kwargs)

        kernel = delta
        support = None
        if len(args) > 0:
            kernel = args[0]
        if len(args) > 1:
            support = args[1]
        if zero_support:
            support = zero

        info1 = engine.get_fes_info(self.fes1)
        emesh1_idx = info1['emesh_idx']
        info2 = engine.get_fes_info(self.fes2)
        emesh2_idx = info2['emesh_idx']

        assert emesh1_idx == emesh2_idx, "convolution is performed only on the same mesh"

        dim1 = self.fes1.GetMesh().Dimension()
        if dim1 == 1:
            if (not self.fes1.FEColl().Name().startswith('L2') and
                    not self.fes1.FEColl().Name().startswith('H1')):
                assert False, "trial(domain) should be H1/L2"
            if (not self.fes2.FEColl().Name().startswith('L2') and
                    not self.fes2.FEColl().Name().startswith('H1')):
                assert False, "test(range) should be H1/L2"
            func = convolve1d
        elif dim1 == 2:
            func = convolve2d
        elif dim1 == 3:
            assert False, "unsupported dimension"
        else:
            assert False, "unsupported dimension"

        #
        # coefficient (we handle constant coefficient separately)
        #
        c_coeff = self._c_coeff
        from petram.phys.phys_model import PhysConstant
        coeff = 0.
        function_coeff = [None, None]
        if c_coeff[0] is not None:
            if isinstance(c_coeff[0], PhysConstant):
                coeff += c_coeff[0].value
            else:
                function_coeff[0] = c_coeff[0]

        if c_coeff[1] is not None:
            if isinstance(c_coeff[1], PhysConstant) and function_coeff[0] is None:
                coeff += 1j*c_coeff[1].value
            else:
                function_coeff[1] = c_coeff[1]

        if (function_coeff[0] is not None) or (function_coeff[1] is not None):
            def eval_coeff(x):
                value = 0
                if function_coeff[0] is not None:
                    value = value + function_coeff[0].EvalValue(x)
                if function_coeff[1] is not None:
                    value = value + 1j*function_coeff[1].EvalValue(x)
                return value
        else:
            eval_coeff = None

        if len(args) != 0:
            self._coeff = (kernel, support)

        M = func(self.fes1,
                 self.fes2,
                 kernel=kernel,
                 support=support,
                 is_complex=is_complex,
                 orderinc=orderinc,
                 trial_domain=trial_domain,
                 test_domain=test_domain,
                 verbose=verbose,
                 coeff=eval_coeff)

        if eval_coeff is None:
            M = M*coeff

        return M



from __future__ import print_function

import numpy as np
import weakref
import scipy
import six

from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD

from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem

from petram.sol.bdr_nodal_evaluator import process_iverts2nodals, eval_at_nodals

def pg(self, attrs, plane = None):                
    '''

    '''
    self.vertices = None
    self.iverts = None
    
    self.knowns = WKD()
    mesh = self.mesh()
    self.iverts = []
    self.attrs = attrs
    self.plane = plane

    attrs = self.attrs
    axyz = self.plane[:3]
    c    = self.plane[-1]

    attr = mesh.GetAttributeArray()
    x = [np.where(attr == a)[0] for a in attrs]
    if np.sum([len(xx) for xx in x]) == 0: return

    ialleles = np.hstack(x).astype(int).flatten()
    ieles = []
    num_tri = 0
    tri_iverts = []
    f_values = []
    for iel in ialleles:
        verts = np.vstack([mesh.GetVertexArray(i)
                           for i in mesh.GetElement(iel).GetVerticesArray()])
        f = np.sum(verts*axyz, -1) + c
        ips = np.where(f > 0)[0]
        ims = np.where(f < 0)[0]
        izs = np.where(f == 0)[0]           
        if (len(ips) == 0 or len(ims) == 0) and len(izs) != 3:
            continue # outside the plane
        ieles.append(iel)
        f_values.append(f)
        
        if f.prod() > 0: # cross section is quad.
           num_tri += 2
        else:
           num_tri += 1

    # then get unique set of elements relating to the verts.
    if num_tri == 0:
        print("not found")
        return 
    vert2el = mesh.GetVertexToElementTable()
           
    iverts = np.array([mesh.GetElement(iel).GetVerticesArray()
                          for iel in ieles])
    self.iverts = iverts
    
    data = process_iverts2nodals(mesh, iverts)
    for k in six.iterkeys(data):
        setattr(self, k, data[k])

    iverts_f = self.iverts_f
    
    # how to interpolate nodal values...
    vertices = np.zeros((num_tri, 3, 3))
    mat      = scipy.sparse.lil_matrix((num_tri*3, len(iverts_f)))

    def fill_vert_mat(vertics, mat, ivv, itri, k, i1, i2, verts, f):
        v1 = verts[i1]; v2 = verts[i2]
        f1 = np.abs(f[i1]); f2 = np.abs(f[i2])
        v = (v1*f2 + v2*f1)/(f1 + f2)
        vertices[itri, k, :] = v
        mat[itri*3+k, ivv[i1]] = np.abs(f1)
        mat[itri*3+k, ivv[i2]] = np.abs(f2)
    
    itri = 0
    for iel, f in zip(ieles, f_values):
        iv = mesh.GetElement(iel).GetVerticesArray()
        verts = np.vstack([mesh.GetVertexArray(i) for i in iv])
        ivv = [np.where(iverts_f == i)[0] for i in iv]
        ips = np.where(f > 0)[0]
        ims = np.where(f < 0)[0]           
        if (len(ips) == 2 and len(ims) == 2):
            fill_vert_mat(vertices, mat, ivv, itri, 0, ips[0], ims[0], verts, f)
            fill_vert_mat(vertices, mat, ivv, itri, 1, ips[0], ims[1], verts, f)
            fill_vert_mat(vertices, mat, ivv, itri, 2, ips[1], ims[0], verts, f)            
            itri += 1

            fill_vert_mat(vertices, mat, ivv, itri, 0, ips[1], ims[0], verts, f)
            fill_vert_mat(vertices, mat, ivv, itri, 1, ips[1], ims[1], verts, f)
            fill_vert_mat(vertices, mat, ivv, itri, 2, ips[0], ims[1], verts, f)            
            itri += 1
        else:

            if len(ips) == 1:
                ims = np.where(f <= 0)[0]                                       
                i =  ips[0]
                ii = ims
            else:
                ips = np.where(f >= 0)[0]   
                i = ims[0]
                ii = ips

            fill_vert_mat(vertices, mat, ivv, itri, 0, i, ii[0], verts, f)
            fill_vert_mat(vertices, mat, ivv, itri, 1, i, ii[1], verts, f)
            fill_vert_mat(vertices, mat, ivv, itri, 2, i, ii[2], verts, f)            
            itri += 1

    self.ibeles = None # can not use boundary variable in this evaulator        
    self.vertices = vertices
    self.interp_mat = mat
    
def eval(self, expr, solvars, phys):
    if self.vertices is None: return None, None
    
    val = eval_at_nodals(self, expr, solvars, phys)

    if val is None: return None, None
    value = self.interp_mat.dot(val)
    value = value.reshape(len(self.vertices), -1)
    return self.vertices,  value

def edge_detect(index):
    print("edge_detect", index.shape)
    store = []
    def check_pair(store, a, b):
        a1 = min(a,b)
        b1 = max(a,b)
        p = (a1, b1)
        if p in store: store.remove(p)
        else: store.append(p)
        return store
        
    for iv in index:
        store = check_pair(store, iv[0], iv[1])
        store = check_pair(store, iv[0], iv[2])
        store = check_pair(store, iv[1], iv[2])        
    ret = np.vstack(store)
    return  ret 
    
        
    
